from __future__ import annotations

import typing

from itertools import chain

from .typing import T_EnumFlag, T


class EnumMatcher:
    """
    Used to compare tuples of :class:`~enum.IntFlag` enums. Tuples are considered 'matching' if
    all flags from the query enums are also set in the base enum.

    Example:
        Consider the enum ::

            >>> import enum
            >>> class Flags(enum.IntFlag):
            ...     ONE = enum.auto()
            ...     TWO = enum.auto()
            ...     COMBINED = ONE | TWO

        Because how :class:`~enum.IntFlag` enums work, we now have
        ``Flags.ONE == int(b'01') == 1`` and ``Flags.TWO == int(b'10') == 2``
        and ``Flags.COMBINED == int(b'11') == 3``. Since all flags in the ``COMBINED``
        value are set, we have

            >>> from ubii.framework.util import EnumMatcher
            >>> EnumMatcher.flags_match(base=Flags.ONE, query=Flags.COMBINED)
            True
            >>> EnumMatcher.flags_match(base=Flags.COMBINED, query=Flags.ONE)
            False

        And for tuples of values ::

            >>> values = (Flags.ONE, Flags.ONE, Flags.TWO, Flags.COMBINED)
            >>> EnumMatcher.matches(base=values, query=tuple([Flags.COMBINED] * len(values)))
            True

        If we have a mapping of :class:`~enum.IntFlag` tuples to some values, we can extract values for 'matching'
        keys:

            >>> mapping = {
            ...     (Flags.ONE, Flags.TWO): "Foo",
            ...     (Flags.TWO, Flags.COMBINED): "Bar"
            ... }
            >>> EnumMatcher.get_matching_value(query=(Flags.COMBINED, Flags.COMBINED), mapping=mapping)
            ["Foo", "Bar"]

    """
    _no_default = object()

    @classmethod
    def flags_match(cls, base: T_EnumFlag, query: T_EnumFlag):
        """
        Compute if all flags set in base are also set in query by computing ``base & query == query``
        """
        return base & query == query

    @classmethod
    def matches(cls, base: typing.Sequence[T_EnumFlag], query: typing.Sequence[T_EnumFlag]) -> bool:
        """

        Args:
            base: a tuple of flag enums that need to be matched
            query: a tuple of flag enums

        Returns:
            ``True`` if :meth:`.flags_match` is ``True`` at every position of the tuples.
            ``False`` otherwise or if tuples have different length or if they contain ``None`` values.
        """
        if not len(base) == len(query):
            return False

        if base == query:
            return True

        if any(x is None for x in chain(base, query)):
            return False

        return all(cls.flags_match(base_val, query_val) for base_val, query_val in zip(base, query))

    @classmethod
    def get_matching_value(cls,
                           query: typing.Sequence[T_EnumFlag],
                           default: typing.Any = _no_default,
                           *,
                           mapping: typing.Mapping[typing.Sequence[T_EnumFlag], T]) -> T:
        """
        Args:
            query: tuple used as query for :meth:`.matches`
            default: default value if no matches are found -- optional
            mapping: Mapping :math:`(IntFlag, IntFlag, ...) \\rightarrow Value`, keys are considered ``base`` values for
                :meth:`.matches`

        Returns:
            All values from ``mapping`` where :meth:`EnumMatcher.matches(key, query) <.matches>` is ``True``,
            otherwise ``default`` if set.

        Raises:
            KeyError: if no matches for ``query`` found in ``mapping`` and no ``default`` is given
        """

        matching = [value for enums, value in mapping.items() if cls.matches(enums, query)]
        if len(matching) != 1 and default is EnumMatcher._no_default:
            raise KeyError(f"Found matching values {matching} for query {query}, not exactly one match")

        return matching[0] if matching else default



from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import enum
import functools
import logging
import typing
import warnings
import weakref

import aiohttp

import ubii.proto
from ubii.framework import (
    topics,
    services,
    client,
    constants,
    util,
    processing, debug,
)
from . import connections

log = logging.getLogger(__name__)


class States(enum.IntFlag):
    """
    States of the default Client Protocol
    """
    STARTING = enum.auto()
    CREATED = enum.auto()
    REGISTERED = enum.auto()
    CONNECTED = enum.auto()
    STOPPED = enum.auto()
    HALTED = enum.auto()

    ANY = STARTING | REGISTERED | CONNECTED | STOPPED | HALTED | CREATED


class LegacyProtocol(client.AbstractClientProtocol[States]):
    """
    The standard protocol creates one UbiiClient, registers it, implements all required behaviours and
    device registration as well as handling of processing modules.
    """

    starting_state = States.STARTING
    """
    """
    end_state = States.STOPPED
    """
    """

    __hook_decorators__ = client.AbstractClientProtocol.__hook_decorators__.union([util.log_call()])
    """
    decorators applied to protocol hooks
    """

    @dataclasses.dataclass
    class Context:
        """
        The fields get assigned during the lifetime of the client
        """
        server: ubii.proto.Server | None = None
        """
        The server message as send by the `master node` after the initial ``server_configuration`` service call
        """
        constants: ubii.proto.Constants | None = None
        """
        The constants used by the `master node` (default topics, data types etc.)
        """
        client: client.UbiiClient | None = None
        """
        The client that owns the protocol
        """
        service_connection: connections.AIOHttpRestConnection | None = None
        """
        Usable service connection created in :meth:`~ubii.node.protocol.LegacyProtocol.create_service_map`
        """
        service_map: services.ServiceMap | None = None
        """
        Usable service map created in :meth:`~ubii.node.protocol.LegacyProtocol.create_service_map`
        """
        topic_connection: connections.AIOHttpWebsocketConnection | None = None
        """
        Usable topic connection created in :meth:`~ubii.node.protocol.LegacyProtocol.create_topic_connection`
        """
        register_manager: typing.AsyncContextManager['client.UbiiClient'] | None = None
        """
        async context manager to register and unregister the client, created in 
        :meth:`~ubii.node.protocol.LegacyProtocol.register_client`
        """
        subscription_manager: SubscriptionManager | None = None
        """
        async context manager which implements the subscriptions, created in 
        :meth:`~ubii.node.protocol.LegacyProtocol.implement_subscriptions`
        """
        session_manager: SessionManager | None = None
        """
        async context manager to start and stop sessions, created in 
        :meth:`~ubii.node.protocol.LegacyProtocol.implement_sessions`
        """
        topic_store: topics.TopicStore[topics.BasicTopic] | None = None
        """
        Used to access topics that the client is subscribed to, created in 
        :meth:`~ubii.node.protocol.LegacyProtocol.implement_subscriptions`
        """
        exc_info: typing.Tuple[Exception | None, typing.Type[Exception] | None, typing.Any] | None = None
        """
        Filled when some protocol task raises an exception
        """
        state_change: typing.Tuple[States, States] | None = None
        """
        The state change that triggered the last callback that handled the context
        """

    def __init__(self,
                 config: constants.UbiiConfig = constants.GLOBAL_CONFIG,
                 log: logging.Logger | None = None):

        super().__init__(config, log)
        self.client: client.UbiiClient | None = None
        """
        Explicitly assign the client to this attribute after initialization, otherwise
        a dedicated client will be created during protocol execution
        """
        self._aiohttp_session: aiohttp.ClientSession = connections.aiohttp_session()
        self.task_nursery.push_async_exit(self._aiohttp_session)

    def _reinit_aiohttp_session(self):
        assert self._aiohttp_session.closed
        self._aiohttp_session: aiohttp.ClientSession = connections.aiohttp_session()
        self.task_nursery.push_async_exit(self._aiohttp_session)
        assert not self._aiohttp_session.closed

    @property
    def aiohttp_session(self):
        """
        shared session for `aiohttp` connections
        """
        return self._aiohttp_session

    @util.cached_property
    def context(self) -> LegacyProtocol.Context:
        """
        Returning context object for better typing
        """
        return self.Context()

    @client.AbstractClientProtocol.state.setter
    def state(self, new_state: States):
        """
        Supress Value Errors when in end state, maybe there are still some tasks trying to change states
        after the protocol finished.
        """
        maybe_suppress = (
            contextlib.suppress(ValueError)
            if self.state.value == self.end_state and not debug()
            else contextlib.nullcontext()
        )
        with maybe_suppress:
            client.AbstractClientProtocol.state.fset(self, new_state)

    async def _set_exc_info(self, *exc_info):
        received_exc = exc_info[0] is not None
        self.context.exc_info = exc_info
        await self.state.set(States.HALTED)

        if received_exc:
            raise exc_info[1]

    async def create_service_map(self, context: LegacyProtocol.Context):
        """
        Create a :class:`~ubii.framework.services.ServiceMap` in the context as ``context.service_map``.

        Also initializes

            *   :attr:`context.server <.Context.server>` from :attr:`.config`
            *   :attr:`context.constants <.Context.server>` from :attr:`.config`
            *   :attr:`context.service_connection <.Context.service_connection>` with working
                :class:`ubii.framework.connections.AIOHttpRestConnection` using :attr:`.aiohttp_session`

        """
        context.server = self.config.SERVER
        context.constants = self.config.CONSTANTS

        context.service_connection = connections.AIOHttpRestConnection(self.config.DEFAULT_SERVICE_URL)
        context.service_connection.session = self.aiohttp_session

        def create_service_call(service: ubii.proto.Service):
            """
            Use service connection of context to create a service call
            """
            assert context.service_connection is not None
            service_call = services.ServiceCall(transport=context.service_connection, mapping=service)
            # add exception handling to new class to not change the default behaviour
            services.ServiceCall.register_decorator(
                util.exc_handler_decorator(self._set_exc_info),
                instance=service_call
            )
            return service_call

        context.service_map = services.DefaultServiceMap(
            service_call_factory=create_service_call
        )

    async def update_config(self, context: LegacyProtocol.Context):
        """
        Update the server configuration in the context. After completion of this coroutine

        *   :attr:`context.server <.Context.server>` is a :class:`~ubii.proto.Server` message with
            the configuration of the master node
        *   :attr:`context.constants <.Context.constants>`  is a :class:`~ubii.proto.Constants` message of the default
            constants of the master node
        """
        assert context.service_map is not None
        assert context.constants is not None
        assert context.server is not None

        service_call = context.service_map[context.constants.DEFAULT_TOPICS.SERVICES.SERVER_CONFIG]
        response = await service_call()
        ubii.proto.Server.copy_from(context.server, response.server)
        ubii.proto.Constants.copy_from(
            context.constants,
            ubii.proto.Constants.from_json(response.server.constants_json)
        )

        assert context.service_connection is not None
        # update service connection url for consistency with topic connection:
        ip = context.server.ip_wlan or context.server.ip_ethernet or 'localhost'
        from urllib.parse import urlparse
        parsed = urlparse(context.service_connection.url)
        context.service_connection.url = f"{parsed.scheme}://{ip}:{context.server.port_service_rest}{parsed.path}"

    async def update_services(self, context: LegacyProtocol.Context):
        """
        Update the service map in the context.

            *   :attr:`context.service_map <.Context.service_map>` is able to perform all service calls advertised
                by the master node after this coroutine completes.
        """

        assert context.service_map is not None
        assert context.constants is not None

        response = await context.service_map[context.constants.DEFAULT_TOPICS.SERVICES.SERVICE_LIST]()
        ubii.proto.ServiceList.copy_from(context.service_map, response.service_list)
        context.service_map.cache_clear()

        # make it easier to call the services
        constants_service_map = {
            name.lower(): topic for name, topic
            in ubii.proto.Constants.DefaultTopics.Services.to_dict(context.constants.DEFAULT_TOPICS.SERVICES).items()
        }
        context.service_map.defaults.update(constants_service_map)

    async def create_client(self, context: LegacyProtocol.Context):
        """
        Create a client in the context.

            *   :attr:`context.client <.Context.client>` is a :class:`UbiiClient` using this protocol as
                :attr:`~UbiiClient.protocol`

        The client is also available as :attr:`.client` attribute
        """

        context.client = self.client or client.UbiiClient(protocol=self)
        if not self.client:
            warnings.warn(
                f"Not setting the protocol client can lead to unexpected behaviour. Created default client {context.client}",
                RuntimeWarning
            )

        self.client = context.client
        assert context.client.protocol == self, (f"{context.client} uses a different protocol "
                                                 f"({context.client.protocol}) instead of {self}")

        await self.state.set(States.CREATED)

    def register_client(self, context: LegacyProtocol.Context):
        """
        Needs to return a context manager, to register and unregister the client, but also
        sets the :attr:`context.register_manager <.Context.register_manager>` attribute
        so that this context manager might be used somewhere else
        """

        @contextlib.asynccontextmanager
        async def context_manager(context: LegacyProtocol.Context):
            """
            Entering the context manager registers the client, exiting it unregisters the client
            """
            assert context.service_map is not None
            assert context.constants is not None

            register_service = context.service_map[context.constants.DEFAULT_TOPICS.SERVICES.CLIENT_REGISTRATION]
            deregister_service = context.service_map[context.constants.DEFAULT_TOPICS.SERVICES.CLIENT_DEREGISTRATION]

            response = await register_service(client=context.client)
            ubii.proto.Client.copy_from(context.client, response.client)

            await self.state.set(States.REGISTERED)

            # ----------- setup ---------------
            yield context.client
            # ----------- teardown ------------
            if context.client.id is None:
                return

            result = await deregister_service(client=context.client)
            assert result.success, f"Client did not deregister correctly! Error: {result.error}"

        context.register_manager = context_manager(context)
        return context.register_manager

    async def create_topic_connection(self, context: LegacyProtocol.Context):
        """
        Create a :attr:`context.topic_connection <.Context.topic_connection>` using a
        :class:`ubii.framework.connections.AIOHttpWebSocketConnection` using :attr:`.aiohttp_session`

        The :attr:`.state` will be set to :attr:`States.HALTED` when the topic connection disconnects.
        """
        assert context.client is not None
        assert context.server is not None
        assert context.client.id, f"Client {context.client} from context is not registered"

        ip = context.server.ip_wlan or context.server.ip_ethernet or 'localhost'
        port = context.server.port_topic_data_ws

        topic_connection = connections.AIOHttpWebsocketConnection(url=f'ws://{ip}:{port}')
        topic_connection.session = self.aiohttp_session

        context.topic_connection = topic_connection
        connection_manager = topic_connection.connect(client_id=context.client.id)
        await self.task_nursery.enter_async_context(connection_manager)

        async def set_state_on_disconnect():
            await topic_connection.events.disconnected.wait()
            await self.state.set(States.HALTED)

        self.task_nursery.create_task(set_state_on_disconnect())

    async def implement_client(self, context: LegacyProtocol.Context):
        """
        Implements
            *   `subscription behaviour <implement_subscriptions>`
            *   `publish behaviour <implement_publish>`
            *   `client registration / deregistration <implement_register>`

        :attr:`context.client <.Context.client>` can be awaited after this coroutine is finished,
        to return a fully functional client.
        """
        assert context.client is not None

        context.client[client.Services].service_map = context.service_map
        self.implement_register(context)
        self.implement_publish(context)

    async def implement_sessions(self, context: LegacyProtocol.Context):
        context.session_manager = SessionManager(context.client)
        session_behaviour: SessionManager = await self.task_nursery.enter_async_context(context.session_manager)

        self.client[client.Sessions].start_session = session_behaviour.start_session
        self.client[client.Sessions].stop_session = session_behaviour.stop_session
        self.client[client.Sessions].sessions = session_behaviour.sessions
        self.client[client.Sessions].get_sessions = session_behaviour.get_sessions

    def implement_subscriptions(self, context: LegacyProtocol.Context):
        """
        Implement :class:`ubii.node.Subscriptions` behaviour of :attr:`context.client <.Context.client>`
        """
        assert context.client is not None

        context.topic_store = topics.TopicStore(
            functools.partial(topics.BasicTopic, task_nursery=self.task_nursery)
        )

        context.subscription_manager = SubscriptionManager(context)

        context.client[client.Subscriptions].subscribe_regex = context.subscription_manager.subscribe_regex
        context.client[client.Subscriptions].subscribe_topic = context.subscription_manager.subscribe_topic
        context.client[client.Subscriptions].unsubscribe_regex = context.subscription_manager.unsubscribe_regex
        context.client[client.Subscriptions].unsubscribe_topic = context.subscription_manager.unsubscribe_topic

    def implement_publish(self, context: LegacyProtocol.Context):  # noqa don't care if it could be static
        """
        Implement :class:`ubii.node.Publish` behaviour of :attr:`context.client <.Context.client>`
        """
        assert context.client is not None

        async def publish(*records: typing.Union[ubii.proto.TopicDataRecord, typing.Dict]):
            assert context.topic_connection is not None
            if len(records) < 1:
                raise ValueError(f"Called {publish} without TopicDataRecord message to publish")

            data = (
                ubii.proto.TopicData(topic_data_record=records[0])
                if len(records) == 1 else
                ubii.proto.TopicData(topic_data_record_list={'elements': records})
            )

            await context.topic_connection.send(data)

        context.client[client.Publish].publish = publish

    def implement_register(self, context: LegacyProtocol.Context):  # noqa
        """
        Implement :class:`ubii.node.Register` behaviour of :attr:`context.client <.Context.client>` using the
        :attr:`context.register_manager <.Context.register_manager>`
        """

        assert context.register_manager is not None
        assert context.client is not None
        context.client[client.Register].register = context.register_manager.__aenter__
        context.client[client.Register].deregister = functools.partial(context.register_manager.__aexit__,
                                                                       None, None, None)

    def implement_devices(self, context: LegacyProtocol.Context):
        """
        Implement :class:`ubii.node.Devices` behaviour of :attr:`context.client <.Context.client>`
        """
        assert context.service_map is not None

        register_service = context.service_map.device_registration
        deregister_service = context.service_map.device_deregistration
        pending_devices: typing.Dict[str, ubii.proto.Device] = {}

        async def _maybe_deregister(device_id):
            if device_id in pending_devices:
                await deregister_service(device=pending_devices[device_id])

        async def register(device: ubii.proto.Device):
            assert context.client is not None
            if device.id:
                raise ValueError(f"{device} already has an id, can't register again")

            device.client_id = context.client.id
            response = await register_service(device=device)

            # update and add to client
            ubii.proto.Device.copy_from(device, response.device)
            context.client.devices += [device]

            # deregister later
            pending_devices[device.id] = device
            self.task_nursery.push_async_callback(_maybe_deregister, device.id)

        async def deregister(device):
            if not device.id:
                raise ValueError(f"{device} does not have an id, not registered")

            response = deregister_service(device=device)

            # remove from client and reset id
            context.client.devices.remove(device)
            ubii.proto.Device.copy_from(device, response.device)

            # don't need to deregister later
            del pending_devices[device.id]

        assert context.client is not None
        context.client[client.Devices].register_device = register
        context.client[client.Devices].deregister_device = deregister

    async def on_create(self, context: LegacyProtocol.Context) -> None:
        assert context.client is not None
        self.implement_subscriptions(context)
        await super().on_create(context)
        assert context.subscription_manager is not None
        await self.task_nursery.enter_async_context(context.subscription_manager)

    async def on_registration(self, context: LegacyProtocol.Context):
        """
        Use :meth:`implement_devices` to implement device registration behaviour for the client,
        and :meth:`implement_processing` to implement handling of processing modules.

        Register all devices of the :attr:`context.client <.Context.client>`
        """
        await super().on_registration(context)

        assert context.client is not None
        log.info(f"{context.client.name} registered.")

        # give client some optional behaviour
        self.implement_devices(context)
        await self.implement_processing(context)
        await self.implement_sessions(context)

        # if client has devices, register them
        to_register = [device for device in context.client.devices]
        # clear devices, because they will be re-added on registration
        context.client.devices.clear()

        for device in to_register:
            await context.client.register_device(device)

        await self.state.set(States.CONNECTED)

    async def implement_processing(self, context: LegacyProtocol.Context):
        """
        Implement :class:`ubii.node.RunProcessingModules` behaviour of :attr:`context.client <.Context.client>`
        """

        assert context.client is not None
        module_manager = ProcessingModuleManager(context)

        context.client[client.RunProcessingModules].get_module_instance = (
            module_manager.get_instance
        )

        assert context.constants is not None
        default_topics = context.constants.DEFAULT_TOPICS

        await context.client.implements(client.Subscriptions)
        subscribe_topic = context.client[client.Subscriptions].subscribe_topic
        assert subscribe_topic is not None
        # register callbacks
        await subscribe_topic(default_topics.INFO_TOPICS.START_SESSION).with_callback(module_manager.on_start_session)
        await subscribe_topic(default_topics.INFO_TOPICS.STOP_SESSION).with_callback(module_manager.on_stop_session)

    async def on_halted(self, context: LegacyProtocol.Context):
        """
        When the protocol is in :attr:`States.HALTED` state, we apply some error handling.

        Currently, this callback checks for connection problems (e.g. unavailable `master node`) and
        restarts the protocol if they are fixed.
        """
        log.info(f"Halted {self}")
        missing_attrs = [field.name for field in dataclasses.fields(context)]
        log.debug(f"attribute[s] {', '.join(missing_attrs)} not set in context when {self} halted")

        if context.service_map is None:
            return False

        connection_problems = (lambda: not context.server or len(context.service_map.elements) == 1)
        fixed_problem = False

        if connection_problems():
            while connection_problems():
                try:
                    await self.update_config(context)
                    await self.update_services(context)
                except aiohttp.client.ClientConnectorError:
                    log.warning(f"Master node not available, waiting for server...")
                    await asyncio.sleep(2)

            await self.state.set(self.starting_state)
            fixed_problem = not connection_problems()

        return fixed_problem

    async def on_start(self, context: typing.Any) -> None:
        for field in dataclasses.fields(context):
            setattr(context, field.name, None)

        if self._aiohttp_session.closed:
            self._reinit_aiohttp_session()

        await super().on_start(context)

    # changed callbacks means state changes have to be adjusted
    state_changes = {
        (None, starting_state): on_start,
        (starting_state, States.CREATED): on_create,
        (States.CREATED, States.REGISTERED): on_registration,
        (States.REGISTERED, States.CONNECTED): client.AbstractClientProtocol.on_connect,
        (States.ANY, States.HALTED): on_halted,
        (States.HALTED, starting_state): on_start,
        (States.ANY, end_state): client.AbstractClientProtocol.on_stop,
        (end_state, starting_state): on_start,
    }
    """
    Callbacks for state changes
    """


class LatePMInitProtocol(LegacyProtocol):
    """
    This is the updated version of the :class:`ubii.node.protocol.LegacyProtocol` able to
    load installed processing modules that require a partly initialized client node.
    """

    async def create_client(self, context: LegacyProtocol.Context):
        """
        In addition to the behaviour of :meth:`LegacyProtocol.create_client` also initializes all
        processing modules of the client defined by it's :class:`InitProcessingModules` behaviour
        """
        await super().create_client(context)
        already_initialized_pms = {module.name: module for module in context.client.processing_modules}

        if context.client.implements(client.InitProcessingModules):
            for name, pm in context.client[client.InitProcessingModules].module_factories.items():
                instance: processing.ProcessingRoutine = pm(context)
                instance.name = name
                assert instance.name in processing.ProcessingRoutine.registry
                already_initialized_pms[instance.name] = instance

        context.client.processing_modules = list(already_initialized_pms.values())


class ResettableProtocol(LatePMInitProtocol):

    async def on_reset(self, context: LegacyProtocol.Context):
        assert context.state_change == (States.STOPPED, None)

    state_changes = LegacyProtocol.state_changes
    state_changes[(States.STOPPED, None)] = on_reset


def __getattr__(name):
    try:
        import importlib.metadata as importlib_metadata
    except ImportError:
        import importlib_metadata

    metadata = importlib_metadata.metadata('ubii-node-python')
    url = metadata['Home-Page']

    if name == 'DefaultProtocol':
        warnings.warn(f"The default Protocol was updated, if you experience bugs with the new DefaultProtocol"
                      f" please use the ``LegacyProtocol`` instead and report the bug at {url}.", DeprecationWarning)

        return ResettableProtocol

    else:
        raise AttributeError(f"{__name__} has no attribute {name}")


class SessionManager:
    """
    This implements the Session behaviour using a context manager, so that topic
    subscriptions are removed when the client stops. Also when the context exit is caused
    by an exception, started sessions are stopped.
    """

    def __init__(self,
                 client: client.UbiiClient,
                 timeout=3,
                 __remove_session_on_clean_exit: bool = False):
        self._client = client
        self.sessions: typing.Dict[str, ubii.proto.Session] = {}
        """
        The started sessions, mapping :math:`id \\rightarrow Session`
        """
        self._start_info: topics.Topic | None = None
        self._stop_info: topics.Topic | None = None
        self._timeout = timeout
        self.stop_sessions_on_exit = __remove_session_on_clean_exit

    async def __aenter__(self) -> SessionManager:
        """
        Subscribes to the info topics for broker communication,
        then returns reference to self
        """
        if self._start_info and self._stop_info:
            return self

        consts: ubii.proto.Constants | None = getattr(self._client.protocol.context, 'constants')
        if not consts:
            raise ValueError(f"No broker constants in client context")

        self._start_info, = await self._client[client.Subscriptions].subscribe_topic(
            consts.DEFAULT_TOPICS.INFO_TOPICS.START_SESSION
        )
        self._stop_info, = await self._client[client.Subscriptions].subscribe_topic(
            consts.DEFAULT_TOPICS.INFO_TOPICS.STOP_SESSION
        )

        return self

    async def __aexit__(self, *exc_info) -> None:
        """
        Remove started sessions if the session manager is closed with
        an exception or its :attr:`.stop_sessions_on_exit` attribute is True

        Args:
            *exc_info: Exception info

        Returns:
            always returns None so possible exceptions are not suppressed
        """
        if not any(exc_info) and not self.stop_sessions_on_exit:
            return

        removed = await asyncio.gather(*map(self.stop_session, self.sessions.values()))
        if not all(removed) and not self.sessions:
            raise ValueError(f"Failed to stop sessions with id[s] {set(self.sessions).difference(set(removed))}")

        log.debug(f"Stopped all remaining sessions {removed}")

        assert not self.sessions

    async def get_sessions(self) -> ubii.proto.SessionList:
        """
        Simple wrapper around a service request which returns the session list inside
        the response

        Returns:
            list of running sessions
        """

        assert self._client.implements(client.Services)
        service = self._client[client.Services].service_map.session_runtime_get_list
        if not service:
            raise ValueError(f"Get Session Service not available")
        result = await service()
        return result.session_list

    async def start_session(self, session: ubii.proto.Session) -> ubii.proto.Session:
        """
        Sends session start request and waits for the broker to confirm start

        Args:
            session: Session specification

        Returns:
            the specifications of the started session as communicated by the broker
        """
        if self._start_info is None:
            raise ValueError(f"Context manager not entered!")

        if session.id:
            raise ValueError(f"Session {session} already started.")

        assert self._client.implements(client.Services)

        info = self._client.task_nursery.create_task(self._start_info.buffer.get())
        response = await self._client[client.Services].service_map.session_runtime_start(session=session)
        informed: ubii.proto.TopicDataRecord = await asyncio.wait_for(info, timeout=self._timeout)
        if informed.session.id != response.session.id:
            raise ValueError(f"Session from info topic not requested session")

        assert response.session.id not in self.sessions
        self.sessions[response.session.id] = response.session

        own_modules = [module for module in informed.session.processing_modules if module.node_id == self._client.id]
        if own_modules:
            assert self._client.implements(client.RunProcessingModules), (
                f"{client} can't run processing modules {own_modules}"
            )

            created = await asyncio.gather(*[
                self._client[client.RunProcessingModules].get_module_instance(
                    module.name,
                    ubii.proto.ProcessingModule.Status.CREATED
                )
                for module in own_modules
            ])

            assert len(created) == len(own_modules)

        return response.session

    async def stop_session(self, session: ubii.proto.Session) -> str:
        """
        Sends a session stop request, waits for the broker to confirm stopping of the session,
        then waits for all associated processing modules to be destroyed.

        Args:
            session: session specification as protobuf message

        Returns:
            the id of the stopped session
        """
        if not session.id:
            raise ValueError(f"Session {session} has no id")

        if session.id not in self.sessions:
            warnings.warn(f"stopping session {session.id} which is not managed")

        assert self._client.implements(client.Services)
        info = self._client.task_nursery.create_task(self._stop_info.buffer.get())

        own_modules = [module for module in session.processing_modules if module.node_id == self._client.id]
        if own_modules:
            stopped = asyncio.gather(*[
                self._client[client.RunProcessingModules].get_module_instance(module.name, module.Status.DESTROYED)
                for module in own_modules
            ])
        else:
            stopped = []

        await self._client[client.Services].service_map.session_runtime_stop(session=session)

        informed: ubii.proto.TopicDataRecord = await asyncio.wait_for(info, timeout=self._timeout)
        if informed.session.id != session.id:
            raise ValueError(f"Session from info topic not requested session")

        if stopped:
            await stopped
            assert len(stopped.result()) == len(own_modules), "Failed to stop own modules"

        removed = self.sessions.pop(session.id)
        log.debug(f"Stopped session {removed}")
        return session.id


class ProcessingModuleManager:
    """
    Handles processing module setup and teardown for a :class:`LegacyProtocol`
    """

    class access_handler:
        def __init__(self, container: typing.Mapping[str, ubii.proto.ProcessingModule]):
            self.container: typing.Mapping[str, ubii.proto.ProcessingModule] = container
            """
            managed container of :math:`name \\rightarrow ProcessingModule`
            """

        def status(self, *statuses: ubii.proto.ProcessingModule.Status) -> typing.List[ubii.proto.ProcessingModule]:
            """
            Get modules with matching status

            Args:
                statuses: allowed protobuf status, only modules with matching status will be returned

            Returns:
                matching processing modules
            """
            return [module for module in self.container.values() if module.status in statuses]

        def name(self, name: str) -> ubii.proto.ProcessingModule | None:
            """
            Get module with matching name. Only one module can match, so it is returned if found

            Args:
                name: a protobuf name, only module with this name will be returned

            Returns:
                processing module if found or None
            """
            return self.container.get(name)

        def id(self, id: str) -> ubii.proto.ProcessingModule | None:
            """
            Get module with matching id. Only one module can match, so it is returned if found

            Args:
                id: a protobuf id, only module with this id will be returned

            Returns:
                generator yielding processing modules
            """
            return {module.id: module for module in self.container.values()}.get(id)

    def __init__(self, context):
        """
        Create a handler for a protocol context
        Args:
            context: the context we are working with
        """
        self.managed_modules: typing.MutableMapping[str, processing.ProcessingRoutine] = {}
        """
        modules instantiated by the manager 
        """
        self.by: ProcessingModuleManager.access_handler = self.access_handler(self.managed_modules)
        """
        Use the specific methods to access the managed modules by status / name / id
        """
        self.pm_subscriptions: weakref.WeakValueDictionary[int, topics.Topic] = weakref.WeakValueDictionary()

        self._context: LegacyProtocol.Context = context
        self._create_specs = {module.name: module for module in context.client.processing_modules}
        self._create_specs_by = self.access_handler(self._create_specs)

        self._managed_modules_changed = asyncio.Condition()

    @contextlib.asynccontextmanager
    async def _handle(self, module: processing.ProcessingRoutine):
        started = await processing.ProcessingRoutine.start(module)
        yield started
        if started.status != started.Status.DESTROYED:
            await processing.ProcessingRoutine.halt(started)

    async def get_instance(
            self,
            name: str,
            *status: ubii.proto.ProcessingModule.Status
    ) -> processing.ProcessingRoutine:
        async with self._managed_modules_changed:
            module: processing.ProcessingRoutine = (
                await self._managed_modules_changed.wait_for(functools.partial(self.managed_modules.get, name))
            )

        async with module.change_specs:
            every_status = set(module.Status)
            await module.change_specs.wait_for(
                lambda: module.status in status or every_status
            )

        return module

    async def create_module(self, name: str) -> None:
        """
        Create a module if specifications for the module are part of the clients processing modules

        Args:
            name: The name of the module

        Returns:
            a processing module instance
        """
        specs = self._create_specs_by.name(name)
        if not specs:
            raise ValueError(f"No module specs for name {name} found")

        factory = None

        if self._context.client.implements(client.InitProcessingModules):
            factory = self._context.client[client.InitProcessingModules].module_factories.get(name)
            if factory:
                factory = functools.partial(factory, self._context)

        factory = factory or processing.ProcessingRoutine

        module: processing.ProcessingRoutine = processing.ProcessingRoutine.registry.get(specs.name)

        if module is None:
            module = factory(**type(specs).to_dict(specs))
            log.debug(f"Created module new with name {specs.name}")
        elif module.status in [module.Status.HALTED, module.Status.DESTROYED]:
            module = factory(**type(specs).to_dict(specs))
            log.debug(f"Created new module of type {type(module)} with name {specs.name}")
        elif type(module).to_dict(module) != type(specs).to_dict(specs):
            module = factory(**type(specs).to_dict(specs))
            log.debug(f"Created new module of type {type(module)} with name {specs.name}")
        else:
            log.warning(f"Already running module with same specifications and name {specs.name}")

        async with self._managed_modules_changed:
            self.managed_modules[name] = module
            self._managed_modules_changed.notify_all()

    async def create_modules_for_sessions(self, session: ubii.proto.Session):
        created = []
        for specs in session.processing_modules:
            if specs.node_id != self._context.client.id:
                continue

            if specs.name not in {pm.name for pm in self._context.client.processing_modules}:
                continue

            if 'name' not in specs:
                raise ValueError(f"Module specs {specs} don't include a name, can't create module without name")

            self._create_specs[specs.name] = specs
            await self.create_module(specs.name)
            module = self.managed_modules[specs.name]

            await processing.ProcessingRoutine.mutate_pm(module, specs)
            await self._context.client.task_nursery.enter_async_context(self._handle(module))
            assert specs.name in self.managed_modules
            created.append(module)

        return created

    async def notify_broker(self,
                            modules: typing.Iterable[ubii.proto.ProcessingModule],
                            remove=False):
        if not modules:
            return

        await self._context.client.implements(client.Services)
        if not remove:
            change_pm_runtime = self._context.client[client.Services].service_map.pm_runtime_add
        else:
            change_pm_runtime = self._context.client[client.Services].service_map.pm_runtime_remove

        await change_pm_runtime(
            processing_module_list={
                'elements': modules
            }
        )

        log.debug(f"Sucessfully {'removed' if remove else 'started'} processing modules {modules}")

    async def handle_io_specs(self, session: ubii.proto.Session):

        # apply io mappings
        for io_mapping in session.io_mappings:
            assert io_mapping.processing_module_id is not None
            assert io_mapping.processing_module_name is not None

            instance = self.by.name(io_mapping.processing_module_name)

            if not instance or instance.node_id != self._context.client.id:
                continue

            assert self._context.topic_store is not None

            await instance.apply_io_mapping(
                io_mapping,
                remote_topic_map=self._context.topic_store
            )

            assert self._context.client.implements(client.Publish)
            assert self._context.client.implements(client.Subscriptions)

            # publish
            if io_mapping.output_mappings:
                topic_map = instance.local_output_topics
                # the local topic map needs to auto-publish data that is written to topics
                decorator = topic_map.on_create_register_callback(self._context.client[client.Publish].publish)
                if decorator not in type(topic_map).default_factory.decorators(topic_map):
                    type(topic_map).default_factory.register_decorator(
                        decorator, instance=topic_map
                    )

            # subscribe
            for mapping in io_mapping.input_mappings or ():
                if mapping.topic_mux:
                    subscribe = self._context.client[client.Subscriptions].subscribe_regex(
                        mapping.topic_mux.topic_selector
                    )
                elif mapping.topic:
                    subscribe = self._context.client[client.Subscriptions].subscribe_topic(
                        mapping.topic
                    )
                else:
                    subscribe = None

                if subscribe:
                    topic, = await subscribe
                    assert topic.on_subscribers_change
                    weakref.finalize(instance, lambda pm_: self.pm_subscriptions.pop(id(pm_), None), instance)
                    self.pm_subscriptions[id(instance)] = topic

            async with instance.change_specs:
                instance.change_specs.notify_all()

    async def on_start_session(self, record: ubii.proto.TopicDataRecord):
        """
        When session is started, handle contained processing modules
        """

        async def _wait_for_status(module: processing.ProcessingRoutine, status: ubii.proto.ProcessingModule.Status):
            async with module.change_specs:
                await module.change_specs.wait_for(lambda: module.status == status)

        session = record.session
        added = await self.create_modules_for_sessions(session)
        await self.handle_io_specs(session)

        await asyncio.gather(
            *(_wait_for_status(module, ubii.proto.ProcessingModule.Status.CREATED) for module in added)
        )
        await self.notify_broker(added)

    async def on_stop_session(self, record):
        """
        Stop processing modules for session.

        Warning:

            When a client subscribes to the stop session topic, it will get sent the last stop request,
            without having started any modules first, we need to handle that.

        """
        session: ubii.proto.Session = record.session
        status = ubii.proto.ProcessingModule.Status
        to_stop = []

        for module in self.by.status(status.CREATED, status.PROCESSING, status.INITIALIZED):
            if module.session_id != session.id:
                continue

            await processing.ProcessingRoutine.halt(module)

        for module in self.by.status(status.HALTED):
            if module.session_id != session.id:
                continue

            to_stop.append(module)
            input_topic_source = self.pm_subscriptions.get(id(module), None)
            if input_topic_source:
                await input_topic_source.remove_subscriber()

        stopped = await asyncio.gather(*map(processing.ProcessingRoutine.stop, to_stop))

        if not stopped == to_stop:
            raise ValueError(f"Could not stop all processing modules from {to_stop}")

        destroyed = [module for module in self.by.status(status.DESTROYED) if module.session_id == session.id]
        async with self._managed_modules_changed:
            for module in destroyed:
                self.managed_modules.pop(module.name)

            self._managed_modules_changed.notify_all()

        await self.notify_broker(destroyed, remove=True)


class SubscriptionManager:
    class OnSubscribersChanged(topics.OnSubscribersChange):
        def __init__(self, topic_store, client_id, as_regex, callback: topics.OnSubscribeCallback):
            super().__init__(client_id, as_regex, callback)
            self.topic_store = topic_store

        def __call__(self, topic: topics.Topic, change: typing.Tuple[int, int]):
            task = super().__call__(topic, change)
            old, new = change
            if new == 0 and old > 0:
                del self.topic_store[topic.pattern]

            return task

    def __init__(self, context: LegacyProtocol.Context):
        self.context: LegacyProtocol.Context = context
        assert context.service_map is not None
        assert context.service_map.topic_subscription
        assert context.topic_store is not None
        self._subscription_handler = functools.partial(self.handle_subscribe, self.context, self.on_subscribe_callback)

    async def on_subscribe_callback(self, client_id, topic: topics.Topic, as_regex: bool = False,
                                    unsubscribe: bool = False):
        message = {
            'client_id': client_id,
            f"{'un' if unsubscribe else ''}"
            f"{'subscribe_topic_regexp' if as_regex else 'subscribe_topics'}": [topic.pattern, ]
        }
        await self.context.service_map.topic_subscription(topic_subscription=message)

    class handle_subscribe(util.CoroutineWrapper):
        """
        Implements the features required by a `subscribe call` according to the Protocol specification
        of the behaviour
        """

        def __init__(self, context, callback, *topic_patterns, as_regex=False, unsubscribe=False):
            self.topic_patterns = topic_patterns
            self.as_regex = as_regex
            self.unsubscribe = unsubscribe
            self.topics = tuple(context.topic_store[topic] for topic in self.topic_patterns)

            assert context.topic_store is not None
            if not self.topic_patterns:
                raise ValueError(f"No topics passed")

            for topic in self.topics:
                if not topic.on_subscribers_change:
                    assert context.client.id
                    topic.on_subscribers_change = SubscriptionManager.OnSubscribersChanged(
                        topic_store=context.topic_store,
                        client_id=context.client.id,
                        as_regex=self.as_regex,
                        callback=callback
                    )

            super().__init__(coroutine=self.__call())

        async def with_callback(self, callback: topics.Consumer) -> (
                typing.Tuple[typing.Tuple[topics.Topic, ...], typing.Tuple]
        ):
            if self.unsubscribe:
                raise NotImplementedError("This is only supported for 'subscribe' calls")

            tokens = [topic.register_callback(callback) for topic in self.topics]
            await asyncio.gather(*[topic.buffer.has_waiting_read for topic in self.topics])
            await self
            return self.topics, tuple(tokens)

        async def __call(self):
            for topic in self.topics:
                if self.unsubscribe:
                    await topic.remove_all_subscribers()
                else:
                    await topic.add_subscriber()

            return self.topics

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc_info):
        await asyncio.gather(*[topic.remove_all_subscribers() for topic in self.context.topic_store.values()])

    def subscribe_regex(self, *pattern: str):
        """
        Subscribe to topic with glob pattern

        Args:
            *pattern: unix wildcard patterns

        Example:

            >>> from ubii.node import *
            >>> client = await connect_client()
            >>> all_infos, = await client[Subscriptions].subscribe_regex('/info/*')

        Returns:
            awaitable returning a tuple of processed topics (one for each pattern, same order)
        """
        return self._subscription_handler(*pattern, as_regex=True, unsubscribe=False)

    def subscribe_topic(self, *pattern: str):
        """
        Subscribe to topic with simple topic name

        Args:
            *pattern: absolute topic names

        Example:

            >>> from ubii.node import *
            >>> client = await connect_client()
            >>> start_pm, = await client[Subscriptions].subscribe_topic('/info/processing_module/start')

        Returns:
            awaitable returning a tuple of processed topics (one for each pattern, same order)
        """
        return self._subscription_handler(*pattern, as_regex=False, unsubscribe=False)

    def unsubscribe_regex(self, *pattern: str):
        """
        Unsubscribe from topic with glob pattern

        Args:
            *pattern: absolute topic names

        Example:

            >>> from ubii.node import *
            >>> client = await connect_client()
            >>> all_infos, = await client[Subscriptions].subscribe_regex('/info/*')
            >>> all_infos.subscriber_count
            1
            >>> all_infos, = await client[Subscriptions].unsubscribe_regex('/info/*')
            >>> all_infos.subscriber_count
            0

        Returns:
            awaitable returning a tuple of processed topics (one for each pattern, same order)
        """
        return self._subscription_handler(*pattern, as_regex=True, unsubscribe=True)

    def unsubscribe_topic(self, *pattern: str):
        """
        Unsubscribe from topic with simple topic name

        Args:
            *pattern: absolute topic names

        Example:

            >>> from ubii.node import *
            >>> client = await connect_client()
            >>> start_pm, = await client[Subscriptions].subscribe_topic('/info/processing_module/start')
            >>> start_pm.subscriber_count
            1
            >>> start_pm, = await client[Subscriptions].unsubscribe_topic('/info/processing_module/start')
            >>> start_pm.subscriber_count
            0

        Returns:
            awaitable returning a tuple of processed topics (one for each pattern, same order)
        """
        return self._subscription_handler(*pattern, as_regex=False, unsubscribe=True)


if typing.TYPE_CHECKING:
    DefaultProtocol = LegacyProtocol
    dynamic_names = ()
else:
    dynamic_names = (
        'DefaultProtocol',
    )

__all__ = dynamic_names + (
    'States',
    'LatePMInitProtocol',
    'LegacyProtocol',
    'SessionManager',
    'ProcessingModuleManager'
)


def __dir__():
    return sorted(__all__)



from __future__ import annotations

import argparse
import collections
import functools
import importlib.resources
import logging.config
import warnings
from typing import (
    Dict,
    List,
)

import sys
import yaml

from . import util

__config__ = yaml.safe_load(importlib.resources.read_text(util, 'logging_config.yaml'))


class _logging_setup:
    log_config = collections.namedtuple('log_config', ['config', 'level', 'warning_filter'])

    @staticmethod
    def add_log_level(name: str, value: int):
        """
        Adds a custom log level to the logging module

        Args:
            value: logging level
            name: name of the new level

        """
        callback_name = name.lower()
        logger_kls = logging.getLoggerClass()
        existing = {
            obj: getattr(obj, attr, None)
            for obj in (logging, logger_kls)
            for attr in (name, callback_name)
        }
        if any(existing.values()):
            warnings.warn(
                f"Can't configure new log level, "
                f"attributes found: {','.join(f'{k} -> {v}' for k,v in existing.items() if v)}"
            )
            return

        assert hasattr(logger_kls, 'log')
        logging.addLevelName(value, name)
        setattr(logging, name, value)
        setattr(logger_kls, callback_name, functools.partialmethod(logger_kls.log, value))
        setattr(logging, callback_name, functools.partial(logging.log, value))

    def __init__(self, base_config=__config__, log_level=logging.INFO, warning_filter: str = 'always'):
        if not hasattr(logging, 'VERBOSE'):
            self.add_log_level('VERBOSE', logging.DEBUG - 5)

        self.base_config = self.log_config(config=base_config, warning_filter=warning_filter, level=log_level)
        self._configs: List[_logging_setup.log_config] = [self.base_config]
        self._applied = False

    @property
    def effective_config(self):
        return self.log_config(**self._configs[-1]._asdict())

    def change(self, config: Dict | None = None, verbosity: int | None = None):
        if config is None and verbosity is None:
            raise ValueError(f"All arguments are None, can' apply change to logging.")

        verbosity = self.effective_config.level if verbosity is None else verbosity

        config = config or {}

        if config.get('incremental', False):
            config = util.merge_dicts(self.effective_config.config, config)
            config['incremental'] = False  # we do the merging, logging framework should just eat the config

        config = config or self.effective_config.config
        config['version'] = 1  # other versions are not supported by the logging framework

        new_config = self.log_config(level=verbosity,
                                     config=config,
                                     warning_filter=self.effective_config.warning_filter)

        if self._applied:
            self._configs.append(new_config)
        else:
            self._configs = [new_config]

        return self

    def _apply(self):
        self._applied = True

        logging.config.dictConfig(self.effective_config.config)
        logging.captureWarnings(True)
        logging.getLogger().setLevel(level=self.effective_config.level)

        if not sys.warnoptions:
            import os, warnings
            warnings.simplefilter(self.effective_config.warning_filter)  # Change the filter in this process
            os.environ["PYTHONWARNINGS"] = self.effective_config.warning_filter  # Also affect subprocesses

    def reset(self):
        self._configs[:] = [self._configs[0]]
        self._apply()

    def __enter__(self):
        self._apply()
        return self

    def __exit__(self, *exc_info):
        config = self._configs.pop()
        if any(exc_info):
            self._configs.clear()

        if self._configs:
            self._apply()
        else:
            self._configs[:] = [config]


logging_setup = _logging_setup(base_config=__config__, log_level=logging.INFO, warning_filter='default')
"""

"""


def parse_args(parser: argparse.ArgumentParser | None = None) -> argparse.Namespace:
    """
    Convenience function to parse command line arguments.
    Adds the following command line arguments to the parser:

        *   ``--verbose``, ``-v`` -- can be specified multiple times, increases the verbosity of the framework logging
        *   ``--debug`` -- sets the framework to :func:`~ubii.framework.debug` mode
        *   ``--log-config`` -- takes a `yaml` file containing a dictionary for :func:`logging.config.dictConfig`
            configuration, an example config is available at
            https://github.com/saggitar/ubii-node-python/blob/develop/src/ubii/framework/util/logging_config.yaml

    Example:

        ::

            from ubii.framework.logging import parse_args
            import argparse

            parser = argparse.ArgumentParser()
            parser.add_argument('--processing-modules', action='append', default=[])
            parser.add_argument('--no-discover', action='store_true', default=False)
            args = parse_args(parser=parser)

    Args:
        parser: if no parser is passed, a new one is created  -- `optional`

    Returns:
        parsed arguments
    """
    parser = parser or argparse.ArgumentParser()

    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--debug', action='store_true', default=False)
    parser.add_argument('--log-config', action='store', default=None)
    args = parser.parse_args()

    verbosity = logging.INFO - 10 * args.verbose
    util.debug(args.debug)

    if util.debug():
        verbosity = min(logging.DEBUG, verbosity)

    if args.log_config:
        with open(args.log_config) as conf:
            args.log_config = yaml.safe_load(conf)
    else:
        args.log_config = __config__

    logging_setup.change(config=args.log_config, verbosity=verbosity)

    return args



"""
Provides the default node implementation as a CLI script in the :mod:`.main` module, and a
helper to discover installed processing modules.
"""
from .main import load_pm_entry_points

__all__ = (
    'load_pm_entry_points',
)



"""
To get specific errors to raise for the different errors the `master node` can send during communication
with a `client node` this module defines some new Exception types that also inherit from :class:`ubii.proto.Error`
and rules for the :class:`proto.marshal.Marshal` for serialization of these Exceptions on the fly.

Importing this module registers the rules for the marshal of the :mod:`ubii.proto` module to convert between
the `Error` protobuf messages (internally used by the :class:`ubii.proto.Error` wrapper) and the new :class:`UbiiError`
exception type i.e. when this module is imported `Error` protobuf messages will not be converted to plain
:class:`ubii.proto.Error` wrapper objects, but instead to :class:`UbiiError` objects so that they can be used in a
:ref:`raise` expression. Depending on e.g. the value if the :attr:`ubii.proto.Error.title` subtypes of
:class:`UbiiError` might be used, i.e. it's possible to differentiate the type of error in an :ref:`except` statement.

Example:

    A short overview of the :doc:`proto-plus marshaling process <plus:marshal>` (since it's not documented in detail):

        Get the marshal name used in your `proto-plus` module, in this case ::

            import ubii.proto
            marshal_name = ubii.proto.__protobuf__.marshal

        Get a reference to the marshal by name (same name returns same object, see documentation of
        :class:`proto.marshal.Marshal`) ::

            import proto.marshal
            marshal = proto.marshal.Marshal(name=marshal_name)

        Create a rule -- you can use the existing rules of the `proto-plus` package as a starting point ::

            from proto.marshal.rules.message import MessageRule
            class CustomRule(MessageRule):
                def to_proto(self, value):
                    # convert wrapper to actual protobuf message type and return it
                    ...

                def to_python(self, value, *, absent: bool | None = None):
                    # convert actual protobuf message to python wrapper and return it,
                    ...

        Register the rule at the Marshal you retrieved earlier. To do that you need the actual protobuf type
        that you want to handle with the rule, if you only got `proto-plus` wrappers at your disposal, you need
        to first get the internal type from the wrapper ::

            Error_pb = ubii.proto.Error.pb()  # actual protobuf type created for the Error wrapper
            marshal.register(Error_pb, CustomRule(descriptor=Error_pb, wrapper=ubii.proto.Error))


See Also:
    :class:`proto.marshal.Marshal` -- documentation of the Marshal class in the `proto-plus` module.

    :obj:`ubii.proto.__protobuf__` -- module level variable of the `ubii-msg-formats` module to expose the `proto-plus`
    marshal and message pool
"""

from __future__ import annotations

import sys
from functools import lru_cache

from ubii.framework import util

if sys.version_info <= (3, 8):
    lru_cache = lru_cache(maxsize=None)

from proto.marshal import Marshal
from proto.marshal.rules.message import MessageRule

import ubii.proto

__protobuf__ = ubii.proto.__protobuf__


@util.dunder.repr('title', 'message', 'stack')
class UbiiError(ubii.proto.Error, Exception, metaclass=ubii.proto.ProtoMeta):
    """
    Base class for all custom errors.
    """

    @classmethod
    @lru_cache
    def rule(cls) -> MessageRule:
        """
        Fallback rule to convert between :class:`ubii.proto.Error` and this type
        """
        return MessageRule(ubii.proto.Error.pb(), cls)

    @property
    def args(self):
        """
        Tuple of (:attr:`.title`, :attr:`.message`, :attr:`.stack`)
        """
        return self.title, self.message, self.stack


class SessionRuntimeStopServiceError(UbiiError):
    """
    Raise if a service request to stop a :class:`ubii.proto.Session` is unsuccessful.
    """


class ErrorRule(MessageRule):
    """
    Custom `MessageRule` to convert `Error` protobuf messages to :class:`UbiiError`
    or :class:`SessionRuntimeStopServiceError` exceptions.
    """

    def to_python(self, value, *, absent: bool | None = None):
        """
        If the :attr:`ubii.proto.Error.title` starts with ``'SessionRuntimeStopService'`` the message is converted to
        a :class:`SessionRuntimeStopServiceError`, otherwise a normal :class:`UbiiError` is created.
        """
        title = value.title or ''
        if title.startswith('SessionRuntimeStopService'):
            return SessionRuntimeStopServiceError.rule().to_python(value, absent=absent)
        else:
            return UbiiError.rule().to_python(value, absent=absent)

    def to_proto(self, value):
        """
        Converts back to a protobuf message using the `MessageRule` implementation from the `proto-plus` package
        """
        return super().to_proto(value)


Marshal(name=__protobuf__.marshal).register(
    ubii.proto.Error.pb(),  # type: ignore
    ErrorRule(ubii.proto.Error.pb(), UbiiError)  # type: ignore
)



from __future__ import annotations

import asyncio
import dataclasses

from ubii.framework.client import UbiiClient
from ubii.node.protocol import DefaultProtocol


class TestClientBehavior:

    async def test_behaviors(self, client, configure_logging):
        await client
        behaviors = client.behaviors
        assert 'optional_behaviors' in behaviors
        assert 'required_behaviors' in behaviors
        assert all(client.implements(b) for b in behaviors['required_behaviors'])

        @dataclasses.dataclass
        class FooBehavior:
            foo: int | None = None

        behaviors['optional_behaviors'] += (FooBehavior,)

        new_protocol: DefaultProtocol = type(client.protocol)()
        new_client = UbiiClient(
            protocol=new_protocol,
            **behaviors
        )
        new_protocol.client = new_client
        new_client = await asyncio.wait_for(new_client, timeout=2)

        assert new_client.id
        assert not new_client.implements(FooBehavior)
        new_client[FooBehavior].foo = 42
        assert new_client.implements(FooBehavior)


"""
This module implements the framework to implement a Ubi Interact Node.
A Ubi Interact Node is expected to perform some communication with the `master node` and conceptualizes
the API for all *master node* interactions, e.g. subscribing to topics, registering devices and so on.
What exactly is expected from a client node to be functional, and what the specific client node actually
implements on top of the required behavior depends on the state of the *master node* implementation.

The currently used *master node* for all Ubi Interact scenarios is the `Node JS node`, which expects
the client node to:

    -   provide an API for all advertised service calls
    -   register itself
    -   establish a data connection to receive :class:`~ubii.proto.TopicData` messages for subscribed topics

Optionally, some client nodes (e.g. a Python Node using the :class:`ubii.node.protocol.DefaultProtocol` or the
`Node JS node` know how to communicate with the `master node` to e.g.

    -   start and stop `Processing Modules`


These two requirements are separated in the python framework:

1.  A :class:`.UbiiClient` defines what kind of behavior and features it is able to implement
2.  A :class:`.AbstractClientProtocol` provides a flexible framework to implement the necessary communication
    with the `master node` to implement those features

A :class:`.UbiiClient` only knows how to communicate with the `master node` through its
:attr:`~.UbiiClient.protocol` i.e. an existing :class:`.UbiiClient` should be able to handle
different `master node` versions by simply using a corresponding protocol if necessary.


A client node typically provides lots of different features, some could be methods (like subscribing
and unsubscribing from topics) or simply other objects that encapsulate different parts of a feature.
Instead of fixing the design, the framework uses :mod:`dataclasses` to describe an arbitrary number of
user defined attributes, grouped by feature:

    -   each :obj:`~dataclasses.dataclass` describes one feature
    -   the :class:`.UbiiClient` is initialized with lists of dataclasses for required and optional behaviors
    -   the attributes of the :obj:`~dataclasses.dataclass` are accessible through the :class:`.UbiiClient` (for
        every class passed during initialization)
    -   since :obj:`dataclasses <dataclasses.dataclass>` enforce the use of typehints, a
        :class:`.UbiiClient` provides a typed API even for its dynamically added attributes (the tradeoff being
        increased verbosity when accessing the attributes)
    -   each attribute will be "assigned" at some point during the execution of the clients
        :attr:`.UbiiClient.protocol`

Note:
    A client is considered usable if all attributes defined by *required* behaviors have been assigned!


By default, the :class:`.UbiiClient` has the required behaviors:
    -   :class:`.Services`
    -   :class:`.Subscriptions`
    -   :class:`.Publish`

and optional behaviors
    -   :class:`.Register`
    -   :class:`.Devices`
    -   :class:`.RunProcessingModules`
    -   :class:`InitProcessingModules`

The :class:`ubii.node.DefaultProtocol` will register the client and implement
the behaviors. The client is considered usable as soon as the required behaviors are implemented,
i.e. when it is able to make service calls, (un-)subscribe to/from topics and publish its own
:class:`ubii.proto.TopicData`.
Subscribing and unsubscribing are technically partly also service calls, but in addition to communicating the
intent to subscribe or unsubscribe to the master node, they return special :class:`~ubii.framework.topics.Topic`
objects, that can be used to handle to published :class:`~ubii.proto.TopicData`. This is explained in
greater detail in the :class:`~ubii.framework.topics.Topic` documentation.
"""

from __future__ import annotations

import abc
import asyncio
import contextlib
import dataclasses
import logging
import typing
import warnings

import itertools

import ubii.proto
from . import (
    services,
    topics,
    util,
    constants,
    protocol, processing
)
from .util.functools import document_decorator
from .util.typing import (
    Protocol,
    T_EnumFlag,
    T,
    Decorator
)

__protobuf__ = ubii.proto.__protobuf__

log = logging.getLogger(__name__)

T_Protocol = typing.TypeVar('T_Protocol', bound='AbstractClientProtocol')

_data_kwargs = {'init': True, 'repr': True, 'eq': True}

BehaviorDict = typing.TypedDict(
    'BehaviorDict',
    {'required_behaviors': typing.Tuple[typing.Type, ...], 'optional_behaviors': typing.Tuple[typing.Type, ...]}
)


class subscribe_call(Protocol):
    patterns: typing.Tuple[str, ...]

    def with_callback(self, callback: topics.Consumer) -> (
            typing.Awaitable[typing.Tuple[typing.Tuple[topics.Topic, ...], typing.Tuple]]
    ):
        """
        set optional callback that should be registered for the subscribed topics

        Args:
            *callback: topic data consumer

        Returns:
            topic and tokens for callback de-registration
        """

    def __call__(self, *pattern: str) -> subscribe_call:
        """
        subscribe_call objects need to have this call signature
        needs to set :attr:`.patterns` attribute.

        Args:
            *pattern: unix wildcard patterns or absolute topic names
        """

    def __await__(self) -> typing.Generator[typing.Any, None, typing.Tuple[topics.Topic, ...]]:
        """
        Returns:
            a tuple of processed topics (one for each pattern, in :attr:`.patterns` same order)
        """


class unsubscribe_call(Protocol):

    def __call__(self, *pattern: str) -> typing.Awaitable[typing.Tuple[topics.Topic, ...]]:
        """
        unsubscribe_call objects need to have this call signature

        Args:
            *pattern: unix wildcard patterns or absolute topic names

        Returns:
            awaitable returning a tuple of processed topics (one for each pattern, same order)
        """


class publish_call(Protocol):
    def __call__(self, *records: ubii.proto.TopicDataRecord | typing.Dict) -> typing.Awaitable[None]:
        """
        publish_call objects need to have this call signature

        Args:
            *records: :class:`~ubii.proto.TopicDataRecord` messages or compatible dictionaries

        Returns:
            some awaitable performing the `master node` communication
        """


class start_session(Protocol):
    async def __call__(self, session: ubii.proto.Session) -> ubii.proto.Session:
        """
        Await to start a session
        Args:
            session: session request to start

        Returns:
            the started session specifications
        """


class stop_session(Protocol):
    async def __call__(self, session: ubii.proto.Session) -> bool:
        """
        Await to stop a session
        Args:
            session: session request to stop

        Returns:
            wether stopping was successful
        """


@dataclasses.dataclass(**_data_kwargs)
class Services:
    """
    Behavior to make service calls (accessed via the service map)

    Example:

        >>> from ubii.node import *
        >>> client = await connect_client()
        >>> assert client.implements(Services)
        >>> await client[Services].service_map.server_config()
        server {
          id: "c2741cca-c75a-41d4-820a-80ac6407d791"
          name: "master-node"
          [...]
        }

    """
    service_map: services.DefaultServiceMap | None = None
    """
    The :class:`~.services.DefaultServiceMap` can be accessed with "shortcuts" for service topics
    
    See Also: 
        :attr:`.services.DefaultServiceMap.defaults` -- how attribute access for the service map works
    """


@dataclasses.dataclass(**_data_kwargs)
class Subscriptions:
    """
    Behavior to subscribe and unsubscribe from topics

    Example:

        >>> from ubii.node import *
        >>> client = await connect_client()
        >>> start_pm, = await client[Subscriptions].subscribe_topic('/info/processing_module/start')
        >>> start_pm.subscriber_count
        1

    """
    subscribe_regex: subscribe_call | None = None
    """
    await to subscribe with regex
    """
    subscribe_topic: subscribe_call | None = None
    """
    await to subscribe with simple topic
    """
    unsubscribe_regex: unsubscribe_call | None = None
    """
    await to unsubscribe with regex
    """
    unsubscribe_topic: unsubscribe_call | None = None
    """
    await to unsubscribe with simple topic
    """


@dataclasses.dataclass(**_data_kwargs)
class Publish:
    """
    Behavior to publish :class:`ubii.proto.TopicDataRecord` messages.
    If multiple records are passed they should be converted to a :class:`ubii.proto.TopicDataList` and published as such,
    otherwise they should be wrapped in a :class:`ubii.proto.TopicData` message.
    """
    publish: publish_call | None = None
    """
    await to publish topic data
    """


@dataclasses.dataclass(**_data_kwargs)
class Register:
    """
    Behavior to optionally unregister and re-register the client node (registering once is probably required
    to establish a data connection for :class:`.Publish` behavior but unregistering and re-registering is typically
    optional -- consult the documentation of the used :class:`protocol <.AbstractClientProtocol>` for details)
    """
    register: typing.Callable[[], typing.Awaitable[UbiiClient]] | None = None
    """
    await to register client node
    """
    deregister: typing.Callable[[], typing.Awaitable[bool | None]] | None = None
    """
    await to unregister client node
    """


@dataclasses.dataclass(**_data_kwargs)
class Devices:
    """
    Behavior to register and deregister Devices (optional)
    """
    register_device: typing.Callable[[ubii.proto.Device], typing.Awaitable[ubii.proto.Device]] | None = None
    """
    await to register a device
    """
    deregister_device: typing.Callable[[ubii.proto.Device], typing.Awaitable[None]] | None = None
    """
    await to deregister a device
    """


@dataclasses.dataclass(**_data_kwargs)
class Sessions:
    """
    Behavior to start and stop Sessions
    """
    sessions: typing.Dict[str, ubii.proto.Session] | None = None
    """
    the sessions started by this node
    """
    start_session: start_session | None = None
    """
    await to start a session 
    """
    stop_session: stop_session | None = None
    """
    await to stop a session 
    """
    get_sessions: typing.Callable[[], typing.Awaitable[ubii.proto.SessionList]] | None = None
    """
    await to get running sessions from broker
    """


class wait_for_module(Protocol):
    def __call__(
            self, name: str,
            *possible_status: ubii.proto.ProcessingModule.Status
    ) -> typing.Awaitable[processing.ProcessingRoutine]:
        """
        Wait until the specified module has the specified status

        Args:
            name: module name
            possible_status: callable should use INITIALIZED by default if no other status is given,
                any statuses given will be accepted

        Returns:
            an awaitable returning the module once it has (one of) the specified status(es)
        """


@dataclasses.dataclass(**_data_kwargs)
class RunProcessingModules:
    """
    Access all running processing module instances
    """
    get_module_instance: wait_for_module | None = None
    """
    Wait until the specified module is e.g. initialized, then return the module
    """


@dataclasses.dataclass(**_data_kwargs)
class InitProcessingModules:
    """
    Behavior to initialize ProcessingModules with custom callables
    """
    module_factories: typing.Mapping[str, typing.Callable[..., processing.ProcessingRoutine]] | None = None
    """
    Mapping :math:`name \\rightarrow factory` for module names to 
    callables which return a :class:`processing.ProcessingRoutine` instance. If the client
    implements it, you can put custom callables inside, so they will get used during module
    instantiation
    """


@util.dunder.repr('id')
class UbiiClient(ubii.proto.Client,
                 typing.Awaitable['UbiiClient'],
                 typing.Generic[T_Protocol],
                 metaclass=util.ProtoRegistry):
    """
    A :class:`UbiiClient` inherits its proto message wrapping capabilities from :class:`ubii.proto.Client`.

    The protocol of the client typically implements the following additional behaviors:

        *   making :class:`ServiceCalls <.services.ServiceCall>` via the :class:`Services` behavior --
            this involves accessing the right Service for your task by topic, and calling it with the right kind of
            data (see https://github.com/SandroWeber/ubi-interact/wiki/Requests for more documentation on default
            topics for services and expected data)

        *   subscribe to topics (or topic patterns) at the master node -- this process involves making the right
            service call and then creating a internal representation of the topic to add callbacks and forward
            received data. Because of this complexity you should not subscribe to topics via a simple `ServiceCall`,
            and instead use the :class:`Subscriptions` behavior. Make sure to use the ``_regex`` version of a method
            when you subscribe to a wildcard pattern.

        *   publish data on topics -- this requires a :class:`~ubii.proto.TopicDataRecord` message or
            a compatible dictionary (see documentation of the message formats) and the :class:`Publish` behavior

        *   run `Processing Modules` -- processing modules need to be registered at the master node.
            Add the modules to the :attr:`~.UbiiClient.processing_modules` field of the client for PMs which can be
            initialized when the client node is created, or to the
            :class:`~InitProcessingModules.late_init_processing_modules` field of the :class:`InitProcessingModules`
            behavior for modules that need to be initialized at a later point of the protocol (e.g. a processing
            module might need to know the *master node's* definition of datatype messages, so it can only be initialized
            after some initial communication between `client` and `master node`.

    The :class:`UbiiClient` will start it's :class:`Client Protocol <AbstractClientProtocol>` when it is awaited
    directly or indirectly (see examples below). The protocol will implement the `behaviors`.

    It's required to link a client and its protocol explicitly::

        from ubii.node.protocol import DefaultProtocol
        from ubii.framework.client import UbiiClient, Services
        import asyncio

        async def main():
            protocol = DefaultProtocol()
            client = UbiiClient(protocol=protocol)
            protocol.client = client

            ...

        asyncio.run(main())

    Awaiting a :class:`UbiiClient` object::

        from ubii.node.protocol import DefaultProtocol
        from ubii.framework.client import UbiiClient, Services
        import asyncio

        async def main():
            protocol = DefaultProtocol()
            client = UbiiClient(protocol=protocol, name="Foo")  # name is a message field
            protocol.client = client
            assert client.name == 'Foo'

            # you could set some attributes before you 'start' the client
            client.is_dedicated_processing_node = True

            # now wait for the client to be usable
            client = await client
            assert client.id  # will be set because the client is registered now

    Using the :class:`UbiiClient` object as an async context manager::

        from ubii.node.protocol import DefaultProtocol
        from ubii.framework.client import UbiiClient, Services
        import asyncio

        async def main():
            protocol = DefaultProtocol()
            client = UbiiClient(protocol=protocol, name="Foo")  # name is a message field
            protocol.client = client

            async with client as running:
                assert running.id  # client is already registered

            assert not client.id  # client gets unregistered when context exits

    When the client is awaited (either directly or as an async context manager) the protocol is started
    internally unless it is already running. Refer to :meth:`.AbstractClientProtocol.start` for details.

    Attributes:
        registry (Dict[str, UbiiClient]): Mapping :math:`id \\rightarrow Client` containing all live :class:`UbiiClients <UbiiClient>`
            with id. Refer to the documentation of :class:`util.ProtoRegistry` for details. ::

                from ubii.framework.client import UbiiClient
                from ubii.node.protocol import DefaultProtocol

                async def main():
                    # you could instead use ubii.node.connect_client
                    protocol = DefaultProtocol()
                    client = UbiiClient(protocol=protocol)
                    protocol.client = client

                    # empty dictionary, since client does not have an id
                    assert not client.id
                    assert not UbiiClient.registry

                    # starts client protocol and returns control when client has id
                    await client

                    assert client.id
                    assert UbiiClient.registry[client.id] == client
    """

    __unique_key_attr__: str = 'id'

    IMPLEMENT_TIMEOUT = None
    """
    Set this value if you want to debug code that hangs in waiting for implementations using
    :attr:`.implements`
    """

    @util.dunder.repr('client')
    class ClientInitTaskWrapper(typing.Awaitable['UbiiClient']):
        """
        This is a wrapper around a task that waits until the client implements the required behaviors,
        and then returns the client. The wrapper can be `reset`, with :attr:`.reset` so that a new
        task is created to be used inside the wrapper.
        """

        def __init__(self, client: UbiiClient):
            self.client = client
            self._task: asyncio.Task | None = None
            self._set_task()

        def _set_task(self):
            self._task = self.client.task_nursery.create_task(self._wait_for_client_implementation())

        def reset(self) -> None:
            """
            Use this method to reset the client behaviors and create a new wrapped task inside the wrapper.

            Returns:
                Reference to self, with new wrapped task

            """
            if not self.client.protocol.finished:
                raise ValueError(
                    f"Can't reset client protocol {self.client}, "
                    f"protocol is not in end state {self.client.protocol.end_state!r}, is in "
                    f"{self.client.protocol.state.value!r} instead."
                )

            for behavior in self.client._behaviors:
                for field in dataclasses.fields(behavior):
                    setattr(self.client[behavior], field.name, None)

            self._set_task()

        async def _wait_for_client_implementation(self):
            await self.client.implements(*self.client._required_behaviors)
            return self.client

        def __await__(self):
            return self._task.__await__()

    def __init__(self: UbiiClient[T_Protocol],
                 mapping=None, *,
                 protocol: T_Protocol,
                 required_behaviors: typing.Tuple[typing.Type, ...] = (
                         Services, Subscriptions, Publish
                 ),
                 optional_behaviors: typing.Tuple[typing.Type, ...] = (
                         Register, Devices, RunProcessingModules, InitProcessingModules, Sessions
                 ),
                 **kwargs):
        """
        Creates a :class:`UbiiClient` object.
        The :class:`UbiiClient` is awaitable. When it is used in an :ref:`await`, the coroutine will
        wait until all attributes for the clients `required_behaviors` are assigned. These assignments
        typically happen as part of the clients :attr:`.protocol` running, sometime the types
        passed as `required_behaviors` or `optional_behaviors` are referred to as
        `behaviors`, and assigning something to their attributes is referred to as `implementing` the behavior.


        Args:
            mapping (Union[dict, ~.Message]): A dictionary or message to be
                used to determine the values for the message fields.
            protocol (AbstractClientProtocol): A concrete protocol instance to be used py the client node
            required_behaviors (typing.Tuple[typing.Type, ...]): tuple of :obj:`~dataclasses.dataclass` types
                that need to be `implemented` by the protocol to consider the `UbiiClient` as usable
            optional_behaviors (typing.Tuple[typing.Type, ...]): tuple of :obj:`~dataclasses.dataclass` types
                that can optionally be `implemented` by the protocol whose attributes can be accessed through the
                `UbiiClient` node.
            **kwargs: passed to :class:`ubii.proto.Client` (e.g. field assignments)
        """
        super().__init__(mapping=mapping, **kwargs)

        self._required_behaviors = required_behaviors or ()
        self._optional_behaviors = optional_behaviors or ()
        behaviors = list(
            itertools.chain(self._required_behaviors, self._optional_behaviors)
        )

        if not all(dataclasses.is_dataclass(b) for b in behaviors):
            raise ValueError(f"Only dataclasses can be passed as behaviors")

        if not self.name:
            self.name = f"Python-Client-{self.__class__.__name__}"  # type: str

        if 'state' not in self:
            self.state = self.State.UNAVAILABLE

        self._change_specs = asyncio.Condition()
        self._notifier = None
        self._protocol = protocol
        self._behaviors = {kls: self._patch_behavior(kls)() for kls in behaviors}

        self._ctx: typing.AsyncContextManager = self._with_running_protocol()
        self._init = self.ClientInitTaskWrapper(self)
        self._init_specs: typing.Dict = type(self).to_dict(self)

    @property
    def initial_specs(self) -> dict:
        """
        Since clients can be reset, the clients current representation needs to be
        separated from the initial protobuf specifications. When the client is :func:`.reset`,
        it's specifications will be set to it's current :attr:`.initial_specs`.

        The initial specs can be adapted during the client's lifetime by
        explicitly assigning to values of this dictionary, otherwise it contains the specifications
        that were used when the object was initialized.
        """
        return self._init_specs

    def _patch_behavior(self, behavior: typing.Type):
        """
        Setting attributes of the behavior should notify the tasks waiting for changed specs of this client,
        e.g. the tasks waiting for implementation state.
        """
        client = self

        # see if (https://github.com/python/mypy/issues/5865) is resolved to check if mypy gets this
        class _(behavior):  # type: ignore
            """
            Proxy to notify client of changed fields.
            """

            def __setattr__(self, key, value):
                super().__setattr__(key, value)
                client.notify()

            def __repr__(self):
                fields = dataclasses.fields(self)
                return (f"{behavior.__module__}.{behavior.__name__}"
                        f"({', '.join('{}={!r}'.format(f.name, getattr(self, f.name)) for f in fields)})")

        from .util.functools import append_doc
        append_doc(_)(behavior.__doc__)
        return _

    def notify(self) -> None:
        """
        Creates a task to notify all coroutines waiting for :attr:`.change_specs` (allows easy notification
        from outside a coroutine i.e. a non-async callback, where it's impossible to acquire the :attr:`.change_specs`
        lock asynchronously)
        """
        assert self.protocol
        assert self._change_specs
        assert hasattr(self, '_notifier')

        async def _notify():
            async with self._change_specs:
                self._change_specs.notify_all()
                self._notifier = None

        if not self._notifier:
            self._notifier = self.task_nursery.create_task(_notify())

    @property
    def task_nursery(self) -> util.TaskNursery:
        """
        the :class:`~codestare.async_utils.nursery.TaskNursery` used by the :attr:`.protocol`
        """
        return self.protocol.task_nursery

    @property
    def change_specs(self) -> asyncio.Condition:
        """
        Allows waiting for behavior attribute assignments.
        See also: :attr:`.implements`
        ::

            from ubii.node import connect_client

            # we use connect_client to create a UbiiClient as well as a protocol and connect them
            # see documentation of connect_client for details

            async def main():
                async with connect_client() as client:
                    await client.change_specs.wait()
                    print("A behavior was implemented!")

        """
        return self._change_specs

    def implements(
            self,
            *behaviors,
            timeout: float | None = IMPLEMENT_TIMEOUT,
    ) -> util.awaitable_predicate:
        """
        Returns an object that can be used to check if the client implements a certain behavior
        or wait until it is implemented.

        ::

            from ubii.node.protocol import DefaultProtocol
            from ubii.framework.client import UbiiClient, Services
            import asyncio

            async def main():
                protocol = DefaultProtocol()
                client = UbiiClient(protocol=protocol)
                protocol.client = client

                async def wait_for_required_behaviors_implicitly():
                    return await client

                async def wait_for_behavior_explicitly():
                    await client.implements(Services)  # used in await expression
                    assert client.implements(Services)  # used in boolean expression

                await asyncio.gather(
                    wait_for_required_behaviors_implicitly(),
                    wait_for_behavior_explicitly()
                )


            asyncio.run(main())

        Args:
            *behaviors: tuple of :obj:`~dataclasses.dataclass` types passed to
                this :class:`UbiiClient` as `required_behaviors` or `optional_behaviors` during initialization.

            timeout: if not None, the returned awaitable will raise a :class:`asyncio.TimeoutError` after specified time
        Returns:
            an :class:`~ubii.framework.util.awaitable_predicate` that converts to
            `True` if all fields of the passed `behaviors` are initialized in this :class:`UbiiClient`
            and / or can be used in an :ref:`await` to wait until that is the case
        """

        def fields_not_none():
            return all(getattr(self._behaviors[b], field.name) is not None
                       for b in behaviors for field in dataclasses.fields(b))

        return util.awaitable_predicate(predicate=fields_not_none, condition=self._change_specs, timeout=timeout)

    @property
    def behaviors(self) -> BehaviorDict:
        """
        Return mapping :math:`(optional / required) \\rightarrow dataclass` showing which behaviors are defined as
        optional or required by this client. You can check their implementation status using :func:`.implements`.
        """
        return {'optional_behaviors': self._optional_behaviors, 'required_behaviors': self._required_behaviors}

    @contextlib.asynccontextmanager
    async def _with_running_protocol(self):
        async with self.protocol:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                client = await self
            yield client

    def __await__(self):
        if self.protocol.finished:
            warnings.warn(f"{self} was stopped, you can "
                          f"try triggering a protocol restart by resetting the "
                          f"client with client.reset()", UserWarning)

        if not self.protocol.was_started:
            self.protocol.start()

        return self._init.__await__()

    def __aenter__(self):
        return self._ctx.__aenter__()

    def __aexit__(self, *exc_info):
        return self._ctx.__aexit__(*exc_info)

    async def reset(self):
        """
        Use this method to reset the client behaviors and allow explicitly restarting the client protocol
        if it is finished. Also resets the protobuf values to the contents of :attr:`.initial_values`

        Warning:

            This behavior is experimental, it is better to simply create a new client instance
        """
        old_id = self.id
        if hasattr(self.task_nursery, 'sentinel_task') and not self.task_nursery.sentinel_task:
            log.debug(f"{self}'s task nursery needs seems to be dead."
                      f" Creating new task nursery for {self}.")

            stack = self.task_nursery.pop_all()
            self.protocol.task_nursery = type(self.task_nursery)(
                name=self.task_nursery.name,
                loop=self.task_nursery.loop
            )
            await self.task_nursery.enter_async_context(stack)

        self._init.reset()

        warning = None

        try:
            await self.protocol.state.set(None)
        except ValueError:
            warning = (f"{type(self.protocol)} does not seem to support resetting the protocol state."
                       f" It does not define a state change from its end state to None!")

        for name, value in self._init_specs.items():
            setattr(self, name, value)

        if warning:
            warnings.warn(warning)
        else:
            log.debug(f"{self} (with old id {old_id!r}) was reset successfully and can be used again.")

    @property
    def protocol(self: UbiiClient[T_Protocol]) -> T_Protocol:
        """
        Reference to protocol used by the client
        """
        return self._protocol

    def __getitem__(self, behavior: typing.Type[T]) -> T:
        return self._behaviors[behavior]

    def __setitem__(self, key, value):
        if not dataclasses.is_dataclass(value):
            raise ValueError(f"can only assign dataclass instances to {key}, got {type(value)}")

        # create copy of value that is a proxy object
        self._behaviors[key] = self._patch_behavior(type(value))(**dataclasses.asdict(value))  # type: ignore
        self.notify()


@util.dunder.all('client')
class AbstractClientProtocol(protocol.AbstractProtocol[T_EnumFlag], util.Registry, abc.ABC):
    """
    :class:`~abc.ABC` to implement client protocols, i.e. define the communication between
    `client node` and `master node` during the lifetime of the client.

    Attributes:
        state_changes: inherited from :class:`~ubii.framework.protocol.AbstractProtocol`
    """
    hook_function: util.registry[str, util.hook] = util.registry(key=lambda h: h.__name__, fn=util.hook)
    """
    This callable wraps the :class:`util.hook <ubii.framework.util.functools.hook>` 
    decorator but registers every decorated function, so that decorators can be easily applied to all registered hooks
    simultaneously
    """
    __hook_decorators__: typing.Set[Decorator] = set()

    @property
    def __registry_key__(self):
        """
        A Standard Protocol can ba associated with _one_ client, so if we have a registered client with
        unique id, we use this id as the key for the registry view.
        During initialisation of the client the key is the default value (~ __module__.__qualname__.#id)
        """
        default = type(self).__default_key_value__
        if not hasattr(self.context, 'client') or not self.context.client or not self.context.client.id:
            return default

        return self.context.client.id

    def __init__(self, config: constants.UbiiConfig = constants.GLOBAL_CONFIG, log: logging.Logger | None = None):
        self.log = log or logging.getLogger(__name__)
        self.config: constants.UbiiConfig = config
        """
        Config used -- contains e.g. default topic for initial `server configuration` service call
        """
        super().__init__()

    @abc.abstractmethod
    async def create_service_map(self, context):
        """
        Create a :class:`~ubii.framework.services.ServiceMap` in the context as
        ``context.service_map`` which has to be able to make a single service call ``context.service_map.server_config``
        """

    @abc.abstractmethod
    async def update_config(self, context):
        """
        Update the server configuration in the context. After completion of this coroutine

            *   ``context.server`` is a :class:`~ubii.proto.Server` message with the configuration of the master node
            *   ``context.constants``  is a :class:`~ubii.proto.Constants` message of the default constants of the
                master node
        """

    @abc.abstractmethod
    async def update_services(self, context):
        """
        Update the service map in the context.

            *   ``context.service_map`` is able to perform all service calls advertised by the master node
                after this coroutine completes.
        """

    @abc.abstractmethod
    async def create_client(self, context):
        """
        Create a client in the context.

            *   ``context.client`` typically is a :class:`ubii.proto.Client` wrapper, e.g. a :class:`UbiiClient`
                which at this moment is not expected to be fully functional.
        """

    @abc.abstractmethod
    def register_client(self, context) -> typing.AsyncContextManager[None]:
        """
        Return a context manager to register the ``context.client`` client, and unregister it when the protocol stops.
        After successful registration the context manager typically needs to also set the protocol state to whatever
        the concrete implementation expects.

            *   ``context.client`` is expected to be up-to-date and usable after registration
        """

    @abc.abstractmethod
    async def create_topic_connection(self, context):
        """
        Should create a :class:`ubii.framework.topics.DataConnection`.

            *   ``context.topic_connection`` is expected to be a fully functional topic connection after
                this coroutine is completed.
        """

    @abc.abstractmethod
    async def implement_client(self, context):
        """
        Make sure the ``context.client`` has fully implemented behavior. The context at this point should contain
        a `context.service_map` and a `context.topic_connection`.

            *   ``context.client`` can be awaited after this coroutine is finished, to return a fully functional client.
        """

    @hook_function
    @document_decorator('.hook_function')
    async def on_start(self, context: typing.Any) -> None:
        """
        Awaits (in order):
            - :meth:`create_service_map`
            - :meth:`update_config`
            - :meth:`update_services`
            - :meth:`create_client`

        The ``context`` is passed for each call, and updated according to the concrete implementation.

        Note:

            For a concrete implementation of a client protocol, assign this callback to a state change in
            :attr:`~ubii.framework.client.AbstractClientProtocol.state_changes`

        Args:
            context: A namespace or dataclass or similar object as container for manipulated values
        """
        await self.create_service_map(context)
        await self.update_config(context)
        await self.update_services(context)
        await self.create_client(context)

    @hook_function
    @document_decorator('.hook_function')
    async def on_create(self, context) -> None:
        """
        Enters the async context manager created by :meth:`register_client` in the :attr:`task_nursery` i.e.
        registers the client and prepares to unregister it if the protocol should be stopped

        The ``context`` is passed to :meth:`register_client`

        Note:
            For a concrete implementation of a client protocol, assign this callback to a state change in
            :attr:`~ubii.framework.client.AbstractClientProtocol.state_changes`

        Args:
            context: A namespace or dataclass or similar object as container for manipulated values
        """
        await self.task_nursery.enter_async_context(self.register_client(context))

    @hook_function
    @document_decorator('.hook_function')
    async def on_registration(self, context) -> None:
        """
        Awaits (in order):
            - :meth:`create_topic_connection`
            - :meth:`implement_client`

        Then the ``context.client`` is awaited to make sure that all behaviors are implemented.
        The ``context`` is passed for each call, and updated according to the concrete implementation.

        Note:

            For a concrete implementation of a client protocol, assign this callback to a state change in
            :attr:`~ubii.framework.client.AbstractClientProtocol.state_changes`

        Args:
            context: A namespace or dataclass or similar object as container for manipulated values

        Raises:
            RuntimeError: if awaiting the ``context.client`` raises a :class:`asyncio.TimeoutError` after a timeout of
                :math:`5s`
        """
        await self.create_topic_connection(context)
        await self.implement_client(context)
        try:
            # make sure client is implemented
            context.client = await asyncio.wait_for(context.client, timeout=5)
        except asyncio.TimeoutError as e:
            raise RuntimeError(f"Client is not implemented") from e

        self.client.state = self.client.State.ACTIVE

    @hook_function
    @document_decorator('.hook_function')
    async def on_connect(self, context) -> None:
        """
        Starts a :class:`ubii.framework.topics.StreamSplitRoutine` in the :attr:`task_nursery` to split
        :class:`ùbii.proto.TopicData` messages from the ``context.topic_connection`` to the topics of the
        ``context.topic_store``

        Note:
            For a concrete implementation of a client protocol, assign this callback to a state change in
            :attr:`~ubii.framework.client.AbstractClientProtocol.state_changes`

        Args:
            context: A namespace or dataclass or similar object as container for manipulated values
        """
        self.task_nursery.create_task(
            topics.StreamSplitRoutine(container=context.topic_store, stream=context.topic_connection)
        )

    @hook_function
    @document_decorator('.hook_function')
    async def on_stop(self, context) -> None:
        """
        Sets the :attr:`~UbiiClient.state` of the :attr:`client` to :attr:`~UbiiClient.State.UNAVAILABLE`

        Note:
            For a concrete implementation of a client protocol, assign this callback to a state change in
            :attr:`~ubii.framework.client.AbstractClientProtocol.state_changes`

        Args:
            context: A namespace or dataclass or similar object as container for manipulated values
        """

        self.log.info(f"Stopped protocol {self}")
        self.client.state = self.client.State.UNAVAILABLE

    def __init_subclass__(cls):
        """
        Register decorators for hook functions
        """
        hook_function: util.hook
        for hook_function, hk in itertools.product(cls.hook_function.registry.values(), cls.__hook_decorators__):
            if hk not in hook_function.decorators():
                hook_function.register_decorator(hk)

        super().__init_subclass__()



"""
The :obj:`ubii.node` package imports the classes from the :mod:`ubii.framework.client` module typically
needed to implement a client node. The :mod:`ubii.node.protocol` submodule implements concrete
:class:`AbstractClientProtocols <ubii.framework.client.AbstractClientProtocol>`, e.g. the
:class:`~ubii.node.protocol.DefaultProtocol` used by the :class:`~ubii.node.connect.connect` 
callable implemented in the :mod:`ubii.node.connect` submodule (imported as
:obj:`connect_client`)

Note:

    You can start a `async` python REPL by using ``python -m asyncio`` to use :ref:`await` directly
    instead of writing everything in a ``main()`` function and using ``asyncio.run()`` (whenever you
    see python interpreter syntax that simply uses `await` assume it's running in an `async` REPL)

Example:

    1.  Make sure the `master node` is running, note the URL for the service endpoint for json data
        e.g. ``http://*:8102/services/json`` (the output of the running master node should show the URL)

    2.  Either set the appropriate environment variable (see :attr:`~ubii.framework.constants.UBII_URL_ENV`) to the
        url of your service endpoint, or pass it to :class:`~ubii.node.connect_client` e.g. ::

            >>> from ubii.node import *
            >>> client = await connect_client('http://localhost:8102/services/json')
            >>> print(client.protocol.state)
            <States.CONNECTED: 8>

    3.  Make sure the client implements the behaviours that you expect or want to try ::
        
            >>> ...
            >>> assert client.implements(Services)
            >>> print(client[Services].service_map)
            {'client_deregistration': <ServiceCall object [topic=/services/client/deregistration]>,
             [...]
             'server_config': <ServiceCall object [topic=/services/server_configuration]>,
             [...]
             'topic_subscription': <ServiceCall object [topic=/services/topic_subscription]>}
            >>> await client[Services].service_map.server_config()
            server {
              id: "c2741cca-c75a-41d4-820a-80ac6407d791"
              name: "master-node"
              [...]
            }
            >>> await client[Services].service_map.client_get_list()
            client_list {
              elements {
                id: "dc3a2252-7501-4b8d-a3dc-cc09be8e98e0"
                name: "Python-Client-UbiiClient"
                [...]
              }
            }
            >>> client.id
            'dc3a2252-7501-4b8d-a3dc-cc09be8e98e0'
            

        You could also inspect which behaviours the client implements and look into their
        documentation. :func:`help` should also give hints about usage of the client behaviours

            >>> client.behaviours
            [<class 'ubii.framework.client.Services'>, 
             <class 'ubii.framework.client.Subscriptions'>, 
             <class 'ubii.framework.client.Publish'>, 
             ...]


See Also:

    :mod:`ubii.framework.client` -- more information how the Ubi Interact python client works


Attributes:

    connect_client: alias for :obj:`ubii.node.connect.connect`
    UbiiClient: alias for :obj:`ubii.framework.client.UbiiClient`
    Services: alias for :obj:`ubii.framework.client.Services`
    Subscriptions: alias for :obj:`ubii.framework.client.Subscriptions`
    Publish: alias for :obj:`ubii.framework.client.Publish`
    Devices: alias for :obj:`ubii.framework.client.Devices`
    Register: alias for :obj:`ubii.framework.client.Register`
    RunProcessingModules: alias for :obj:`ubii.framework.client.RunProcessingModules`
    InitProcessingModules: alias for :obj:`ubii.framework.client.InitProcessingModules`

"""
from ubii.framework.client import (
    UbiiClient,
    Services,
    Subscriptions,
    Publish,
    Devices,
    Register,
    RunProcessingModules,
    InitProcessingModules
)
from .connect import (
    connect as connect_client
)

from .protocol import (
    LegacyProtocol
)

__all__ = (
    'LegacyProtocol',  # noqa
    'connect_client',
    'UbiiClient',
    'Services',
    'Subscriptions',
    'Publish',
    'Devices',
    'Register',
    'RunProcessingModules',
    'InitProcessingModules',
)





"""
Defines some special TypeVars and Type Aliases to be used throughout the codebase.

Also fixes the imports for some names that are not compatible with older python versions, namely
:obj:`.runtime_checkable` and :obj:`.Protocol`, so the special import handling can be done in one place
instead of every module that imports them.


Note:

    *   :obj:`.runtime_checkable` is an alias for :obj:`typing.runtime_checkable` or its backport from
        `typing_extensions`_ for older python interpreter versions

    *   :obj:`.Protocol` is an alias for :obj:`typing.Protocol` or its backport from `typing_extensions`_ for older
        python interpreter versions

    .. _typing_extensions: https://github.com/python/typing/blob/master/typing_extensions/README.rst
"""

from __future__ import annotations

import enum
from typing import (
    TypeVar,
    Coroutine,
    Any,
    Callable,
    Tuple,
    Optional,
    Type, Awaitable
)

try:
    from typing import runtime_checkable, Protocol
except ImportError:
    from typing_extensions import runtime_checkable, Protocol

T = TypeVar('T')
"""
Simple TypeVar
"""
S = TypeVar('S')
"""
Simple TypeVar
"""
R = TypeVar('R')
"""
Simple TypeVar
"""

T_Exception = TypeVar('T_Exception', bound=Exception)
"""
bound to :class:`Exception`
"""
T_EnumFlag = TypeVar('T_EnumFlag', bound=enum.IntFlag)
"""
bound to :class:`enum.IntFlag`
"""

T_co = TypeVar('T_co', covariant=True)
"""
covariant values
"""
T_contra = TypeVar('T_contra', contravariant=True)
"""
contravariant values
"""

SimpleCoroutine = Coroutine[Any, Any, T]
"""
TypeAlias for a Coroutine that only cares about return Type
"""
Decorator = Callable[[Callable], Callable]
"""
A Callable that decorates other Callables
"""
ExcInfo = Tuple[Optional[Type[T_Exception]], Optional[T_Exception], Any]
"""
Types according to return of :func:`sys.exc_info`
"""


@runtime_checkable
class Descriptor(Protocol[T_co]):
    def __get__(self, instance: Any | None = ..., owner: Type[Any] | None = ...) -> T_co:
        """
        Descriptors need to have this method. For more inforation see https://docs.python.org/3/howto/descriptor.html
        """


@runtime_checkable
class Documented(Protocol):
    """
    Attributes:
        __doc__: Documented objects need to have a ``__doc__`` attribute
    """
    __doc__: str


AsyncGetter = Callable[[], Awaitable[T]]
"""
"""

AsyncSetter = Callable[[T], Awaitable[None]]
"""
"""

__all__ = (
    'SimpleCoroutine',
    'Decorator',
    'ExcInfo',
    'Descriptor',
    'Documented',
    'Protocol',
    'runtime_checkable',
    'T',
    'S',
    'R',
    'T_Exception',
    'T_EnumFlag',
    'T_co',
    'T_contra',
    'AsyncSetter',
    'AsyncGetter',
)
"""
names that should be imported from :mod:`ubii.util.typing` and not from other modules
"""



"""
This is a pytest plugin that you can require in your pytest conftest.py
files to easier write python tests for the Ubi-Interact Python node.

More information:
https://docs.pytest.org/en/7.1.x/how-to/writing_plugins.html#requiring-loading-plugins-in-a-test-module-or-conftest-file

Example :

    In your root level `conftest.py` file, require the module like ::

        pytest_plugins = ['ubii.node.pytest']


"""

from __future__ import annotations

import asyncio
import contextlib
import logging
import pathlib
import typing
import typing as t
import warnings

import proto.message
import pytest
import yaml

import codestare.async_utils

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata

import ubii.proto as ub
from ubii.framework.client import UbiiClient, InitProcessingModules
from ubii.framework.logging import logging_setup
from ubii.node.protocol import DefaultProtocol

log = logging.getLogger(__name__)

__verbosity__: int | None = None
_error_marker = 'closes_loop'


def pytest_addoption(parser):
    """
    Adds command line option `--log-config` and .ini option `data_dir`
    """
    parser.addoption(
        "--log-config",
        action="store", default="./data/logging_config.yaml", help="path to yaml file containing log config"
    )
    parser.addini(
        'data_dir',
        default='./data', help='Relative path to directory with test data.'
    )
    parser.addini(
        'cli_entry_point',
        default='ubii-client', help='Entry point for CLI'
    )
    parser.addini(
        'write_test_references',
        default=True, help='Some tests write additional data to the data_dir, set to false to disable'
    )


def pytest_configure(config):
    """
    Sets verbosity and checks if the protobuf package is installed correctly
    """
    global __verbosity__
    __verbosity__ = logging.INFO - 5 * config.getoption('verbose')

    import ubii.proto
    assert ubii.proto.__proto_package__ is not None, "No proto package set, aborting test setup."

    # register an additional marker
    config.addinivalue_line(
        "markers",
        f"{_error_marker}: mark test which close the event loop"
    )


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_setup(item: pytest.Item):
    yield
    marker = item.get_closest_marker(_error_marker)
    loop_scope = {n: v for n, v in item.user_properties}.get('event_loop_scope', None)
    if marker and loop_scope != 'function':
        raise pytest.UsageError(f"{item} is marked as closing the event loop, it needs to request"
                                f" a function scoped event_loop fixture to create a new loop for each call")


@pytest.fixture(autouse=True)
def configure_verbosity(caplog):
    caplog.set_level(__verbosity__)
    yield


@pytest.fixture(scope='session')
def configure_logging(request):
    """
    Change log config if fixture is requested.
    You should use the caplog fixture if possible!

    Args:
        request: will be passed if fixture is parametrized indirectly,
            `request.param` should contain the logging config as dictionary

    """
    from pathlib import Path
    log_config_path = Path(request.config.getoption('--log-config'))
    if log_config_path.exists():
        with log_config_path.open() as f:
            test_logging_config = yaml.safe_load(f)
            logging_setup.change(config=test_logging_config)

    custom = getattr(request, 'param', None)
    logging_setup.change(config=custom, verbosity=__verbosity__)

    with logging_setup:
        yield logging_setup


class TestDataHandler:
    """
    Easier management of test data files
    """

    def __init__(self, dir_path: pathlib.Path):
        self.dir_path = dir_path
        self.dir_path.mkdir(exist_ok=True, parents=True)

    def path(self, filename: str):
        return self.dir_path / filename

    @contextlib.contextmanager
    def write(self, filename: str, mode='w'):
        with self.path(filename).open(mode=mode) as f:
            yield f

    @contextlib.contextmanager
    def read(self, filename: str, mode='r'):
        with self.path(filename).open(mode=mode) as f:
            yield f


@pytest.fixture
def test_data(pytestconfig, data_dir, tmp_path, request) -> TestDataHandler:
    cls = getattr(request.node, 'cls', None)
    if cls:
        data_dir = data_dir / cls.__name__

    yield TestDataHandler(
        data_dir / request.node.callspec.id if pytestconfig.getini('write_test_references') else tmp_path
    )


@pytest.fixture(scope='session', autouse=True)
def service_url_env(request):
    """
    Sets environment variable used for connection to broker node
    to default or requested value
    """
    import os
    from ubii.framework.constants import UBII_URL_ENV
    old = os.environ.get(UBII_URL_ENV)
    os.environ[UBII_URL_ENV] = getattr(request, 'param', None) or 'http://localhost:8102/services/json'
    yield
    if old:
        os.environ[UBII_URL_ENV] = old


@pytest.fixture(autouse=True, scope='session')
def debug_settings():
    """
    Enables debug mode, automatically
    """
    from ubii.framework import debug
    previous = debug()
    debug(enabled=True)
    yield
    debug(enabled=previous)


@pytest.fixture(scope='session')
def data_dir(pytestconfig) -> pathlib.Path:
    """
    Configures data directory and returns it so tests can write stuff to it
    """
    data_dir_config_value = pytestconfig.getini('data_dir')
    data_dir = pytestconfig.rootpath / data_dir_config_value
    assert data_dir.exists(), f"Wrong data dir: {data_dir.resolve()} does not exist."
    yield data_dir


@pytest.fixture(scope='session')
def cli_entry_point(pytestconfig) -> typing.Callable:
    """
    Load entry point for CLI, according to pytest config
    """
    entry_point = pytestconfig.getini('cli_entry_point')

    with warnings.catch_warnings():
        # this deprecation is discussed a lot
        warnings.simplefilter("ignore")
        loaded = [
            entry.load()
            for entry in metadata.entry_points().get('console_scripts', ())
            if entry.name == entry_point
        ]
    assert len(loaded) == 1, (f"{len(loaded)} entry point[s] for specification {entry_point} found in python path. "
                              f"Did you correctly install the [cli] extra?")
    assert callable(loaded[0]), f"Entry point {loaded[0]} is not callable."
    yield loaded[0]


@pytest.fixture(scope='session')
def base_module():
    """
    Returns an empty processing module
    """
    yield ub.ProcessingModule()


@pytest.fixture(scope='session')
def base_session():
    """
    Returns an empty session
    """
    yield ub.Session()


@pytest.fixture(scope='session')
def base_client():
    """
    Returns an empty session
    """
    yield ub.Client()


P = t.TypeVar('P', bound=proto.message.Message)


def _change_specs(message: P, *specs: P):
    base = type(message).pb(message)
    for change in specs:
        if not isinstance(change, proto.message.Message):
            change = type(message)(mapping=change)
        base.MergeFrom(type(change).pb(change))

    type(message).copy_from(message, base)


_get_param = (lambda request: request.param if hasattr(request, 'param') else ())


@pytest.fixture(scope='class')
def late_init_module_spec(request):
    """
    Yield the list of module types specified as the request
    """
    factory = _get_param(request)
    if factory:
        factories = list(request.cls.late_init_module_spec)
        _, _, name = factories[request.param_index]
        yield {name: factory}
    else:
        yield


@pytest.fixture(scope='class')
def module_spec(base_module, request):
    """
    Update the base module with all changes from the request
    """
    _change_specs(base_module, *_get_param(request))
    yield base_module


@pytest.fixture(scope='class')
def session_spec(base_session, module_spec, request):
    """
    Update the base session with all changes from the request
    """
    _change_specs(base_session, *_get_param(request))
    module_names = [pm.name for pm in base_session.processing_modules]

    if module_spec.name and module_spec.name not in module_names:
        base_session.processing_modules += [module_spec]
    yield base_session


@pytest.fixture(scope='class')
async def client_spec(
        base_client,
        module_spec,
        request
):
    """
    Update the base client with all changes from the request
    """
    _change_specs(base_client, *_get_param(request))

    by_name = {pm.name: pm for pm in base_client.processing_modules}
    if module_spec.name in by_name:
        by_name[module_spec.name] = module_spec

    base_client.processing_modules = list(by_name.values())
    yield base_client


@pytest.fixture(scope='class')
async def client(client_spec, late_init_module_spec) -> UbiiClient:
    """
    We need more control over the client, so don't use the default client interface.
    """
    protocol = DefaultProtocol()
    client = UbiiClient(**type(client_spec).to_dict(client_spec), protocol=protocol)
    protocol.client = client
    type(client).IMPLEMENT_TIMEOUT = 2

    if late_init_module_spec:
        client[InitProcessingModules].module_factories = late_init_module_spec

    yield client
    if not client.protocol.finished and client.protocol.was_started:
        await client.protocol.stop()


@pytest.fixture(scope='session')
def event_loop() -> asyncio.AbstractEventLoop:
    """
    We need better control over the asyn processing
    """

    loop = asyncio.get_event_loop_policy().new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop

    tasks = asyncio.all_tasks(loop=loop)
    if tasks:
        for nursery in codestare.async_utils.TaskNursery.registry.values():
            if any(t in nursery.tasks for t in tasks):
                loop.run_until_complete(nursery.__aexit__(None, None, None))

        tasks = asyncio.all_tasks(loop=loop)
        for task in tasks:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                loop.run_until_complete(task)

    loop.close()


def params(name: str, *values):
    return [pytest.param(value, id=f"{name}{value}") for value in values]


def pytest_generate_tests(metafunc):
    """
    Automatically parametrizes tests in classes with class attributes `client_spec`, `module_spec` or
    `session_spec` that use the corresponding fixtures.
    """
    specs = [
        'client_spec',
        'module_spec',
        'session_spec',
        'late_init_module_spec'
    ]

    for spec in specs:
        if hasattr(metafunc.cls, spec) and spec in metafunc.fixturenames:
            parametrization = getattr(metafunc.cls, spec)
            if not isinstance(parametrization, typing.Iterable):
                parametrization = None

            additional_params = getattr(metafunc.cls, f"{spec}_params", None)
            assert not additional_params or parametrization, (
                f"Additional params {additional_params} without parametrization {spec}"
            )

            if parametrization:
                if not additional_params:
                    metafunc.parametrize(spec, parametrization, indirect=[spec])
                    continue

                param_dict = {id_: (values, marks, id_) for values, marks, id_ in parametrization}
                param_ids = list(param_dict)

                for param_id in param_ids:
                    param_string = spec
                    params = additional_params.get(param_id, {})
                    values, marks, id = param_dict.pop(param_id)
                    for n, (new_values, new_marks, new_id) in params.items():
                        param_string += f',{n}'
                        id += f"-{new_id}"
                        values = *values, *new_values
                        marks = *marks, *new_marks

                    param_dict.setdefault(param_string, []).append(pytest.param(*values, marks=marks, id=id))

                for k, params in param_dict.items():
                    metafunc.parametrize(k, params, indirect=[*k.split(',')])



"""
This module provides some classes to implement the handling of the :class:`ubii.proto.TopicData` messages
which are send by the master node when the client subscribes to a topic via a
:class:`ubii.framework.services.ServiceCall`.

The client node will need a :class:`DataConnection` to the *master node*, e.g. a websocket connection. Some example
connections are implemented in :mod:`ubii.framework.connections`. At which point during its lifetime the client is
able to establish this :class:`DataConnection` is described in greater detail in the :mod:`ubii.node.protocol` module.

After the :class:`DataConnection` is established, messages can be read using the async interator API and send
using the :meth:`DataConnection.send` method.

Note:

    The JS API currently defines methods to directly add a callback for a topic pattern -- to do the same in the Python API
    you first get a reference to the specific :class:`Topic` using the pattern, then add the callback with the topics
    :meth:`~Topic.register_callback` method. ::

        import logging
        from ubii.node import connect_client
        from ubii.framework.client import Subscriptions

        log = logging.getLogger('MyLogger')

        async def main():
            async with connect_client() as client:
                topic, = await client[Subscriptions].subscribe_topic('foo/bar') # get a topic reference
                topic.register_callback(log.info) # add callback

See Also:
    :class:`TopicStore` -- details how topics are retrieved by pattern
"""

from __future__ import annotations

import abc
import asyncio
import contextlib
import functools
import logging
import re
import typing
import warnings
import weakref

import proto
import ubii.proto

from . import util
from .util.typing import (
    T_contra,
    Protocol,
)

__protobuf__ = ubii.proto.__protobuf__

T_Buffer = typing.TypeVar('T_Buffer')
T_Token = typing.TypeVar('T_Token')

log = logging.getLogger(__name__)


class Consumer(Protocol[T_contra]):
    def __call__(self, value: T_contra) -> typing.Coroutine[typing.Any, typing.Any, None] | None:
        """
        Consumer objects need to have this call signature

        Args:
            value: something to consume

        Returns:
            no return or coroutine with no return
        """


class DataConnection(typing.AsyncIterator[ubii.proto.TopicData]):
    """
    A DataConnection can be used to asynchronously iterate over received TopicData messages,
    and to send TopicData messages to the master node.
    """

    @abc.abstractmethod
    async def send(self, data: ubii.proto.TopicData): ...


@util.dunder.repr('topic', 'callback')
class TopicCoroutine(util.CoroutineWrapper[typing.Any, typing.Any, None], typing.Generic[T_Buffer]):
    """
    A topic coroutine waits until a value is written to the topic and then runs its callback.
    """

    def __init__(self, *,
                 buffer: typing.AsyncIterator[T_Buffer],
                 callback: Consumer[T_Buffer]):
        """
        Args:
            buffer: Iterator producing buffer value
            callback: a callable consuming the buffer values
        """
        self.__name__ = callback.__name__ if hasattr(callback, '__name__') else repr(callback)  # type: ignore
        self.topic = buffer
        self.callback = util.make_async(callback)
        super().__init__(coroutine=self._run())

    async def _run(self):
        async for value in self.topic:
            log.log(level=5, msg=f"Running {self.callback} on {value}")
            await self.callback(value)


class TopicDataBufferManager(typing.Generic[T_Buffer], abc.ABC):
    """
    Simple :class:`~abc.ABC` to make sure the inheriting class has a buffer field
    """

    @property
    @abc.abstractmethod
    def buffer(self: Topic[T_Buffer, T_Token]) -> util.accessor[T_Buffer]:
        """
        Use this attribute to synchronize access to the managed resource
        """


@util.dunder.repr('pattern', 'token_factory', 'task_nursery')
class Topic(typing.AsyncIterator[T_Buffer], TopicDataBufferManager[T_Buffer], typing.Generic[T_Buffer, T_Token],
            abc.ABC):
    """
    A :class:`Topic` can be used to asynchronously iterate over :class:`TopicDataRecords <ubii.proto.TopicDataRecord>`
    which are published to the topic.
    It can also register (and unregister) callbacks to handle the published values in a background task.

    To "publish" a value to the topic locally use
    :meth:`await topic.buffer.set(...) <codestare.async_utils.descriptor.accessor.set>`.

    The :class:`Client Protocol <ubii.framwork.client.AbstractClientProtocol>` makes sure that all
    :class:`~ubii.proto.TopicDataRecord` messages received via the clients
    :class:`DataConnection` (created in its :attr:`~ubii.framwork.client.UbiiClient.Context`)
    get forwarded to matching topics in this way. One doesn't need to manually "publish" data
    except when e.g. mocking a connection.

    To publish TopicData to the master node instead use the :class:`~ubii.framework.client.Publish` behaviour of the
    :class:`~ubii.framework.client.UbiiClient`.

    Warning:

        The `master node` does not allow the clients to subscribe to the same topic multiple times. The application
        code on the other hand might want to "subscribe" to the topic, without caring about other subscriptions in other
        parts of the application. Implementing this kind of "subscription manager" can be easily done with the
        :attr:`.subscriber_count` property as well as the :meth:`.add_subscriber`, :meth:`remove_subscriber` and
        :meth:`remove_all_subscribers` methods, which trigger the :attr:`.on_subscribers_change` callback.
        E.g. set the actual service call as :attr:`.on_subscribers_change` callback of the topic and simply count
        calls to 'subscribe' and 'unsubscribe' from the topic by setting the virtual :attr:`.subscriber_count`.

    See Also:

        :class:`~ubii.node.protocol.LegacyProtocol.implement_subscriptions` -- example implementation of
        :class:`~ubii.framework.client.Subscriptions` behaviour using this module


    Attributes:
        buffer: inherited from :class:`TopicDataBufferManager`
    """
    on_subscribers_change: OnSubscribersChange | None

    async def __anext__(self) -> T_Buffer:
        value = await self.buffer.get()
        assert value is not None
        return value

    def __init__(self: Topic[T_Buffer, T_Token],
                 pattern,
                 *,
                 token_factory: typing.Callable[[], T_Token],
                 task_nursery: util.TaskNursery | None = None,
                 **kwargs) -> None:
        """

        Args:
            pattern: initializes :attr:`.pattern`
            token_factory: initializes :attr:`.token_factory`
            task_nursery: initializes :attr:`.task_nursery`
            **kwargs: passed to super class init methods
        """
        super().__init__(**kwargs)
        self.on_subscribers_change: OnSubscribersChange | None = None
        """
        Special callback to execute when :attr:`.subscriber_count` changes -- defaults to :obj:`None`
        """
        self.pattern: str = pattern
        """
        Wildcard pattern or simple string defining the topic
        """
        self.task_nursery: util.TaskNursery = task_nursery or util.TaskNursery(name=f"Task Nursery for {self}")
        """
        takes care of managing tasks for callbacks
        """
        self.token_factory: typing.Callable[[], T_Token] = token_factory
        """
        called to create unique tokens to unregister callbacks, see :meth:`.register_callback`
        """
        self.callback_tasks: typing.Dict[T_Token, asyncio.Task] = {}
        """
        Mapping :math:`Token \\rightarrow Task` to access created callback tasks
        """
        self._subscriber_count = 0

        self._callback_tokens: typing.Dict[int, typing.Tuple[T_Token, ...]] = {}

    @property
    def subscriber_count(self):
        """
        number of virtual "subscribers" to this topic. Executes :attr:`.on_subscribers_changed` when it is
        set -- with args

            *   ``topic`` = this topic
            *   ``change`` = (old subscriber count, new subscriber count)
        """
        return self._subscriber_count

    @subscriber_count.setter
    def subscriber_count(self, value):
        if self.on_subscribers_change:
            self.on_subscribers_change(self, (self.subscriber_count, value))

        self._subscriber_count = value

    @contextlib.asynccontextmanager
    async def _wait_for_event(self):
        self.on_subscribers_change.event.clear()
        yield
        await self.on_subscribers_change.event.wait()

    async def add_subscriber(self):
        """
        Increase subscriber count and wait for :attr:`.on_subscriber_change` callback to finish
        """
        async with self._wait_for_event():
            self.subscriber_count += 1

    async def remove_subscriber(self):
        """
        Decrease subscriber count and wait for :attr:`.on_subscriber_change` callback to finish
        """
        async with self._wait_for_event():
            self.subscriber_count -= 1

    @util.cached_property
    def registered_callbacks(self) -> typing.Container[typing.Callable]:
        """
        Use this property to check if a callback is registered in this topic

        Example:
            >>> import itertools
            >>> topic = Topic('foo', token_factory=itertools.count().__next__)
            >>> print in topic.registered_callbacks
            False
            >>> token = topic.register_callback(print)
            >>> print in topic.registered_callbacks
            True
        """

        class id_container:
            def __init__(self, mapping: typing.Mapping[T_Token, typing.Any]):
                self._mapping = mapping

            def __contains__(self, item):
                return id(item) in self._mapping

        return id_container(self._callback_tokens)

    def get_token(self, callback: typing.Callable) -> typing.Tuple[T_Token]:
        """
        If the callback is registered in this topic, get the registration token[s]

        Args:
            callback: some callback that could have been registered previously

        Returns:
            tuple of all registration tokens for callable or None if callback was not registered
        """
        return self._callback_tokens.get(id(callback), None)

    async def remove_all_subscribers(self):
        """
        Set subscriber count to ``0`` and wait for :attr:`.on_subscriber_change` callback to finish
        """
        async with self._wait_for_event():
            self.subscriber_count = 0

    def register_callback(self, callback: Consumer[T_Buffer]) -> T_Token:
        """
        Register a callback function as a callback on this topic.

        This starts a task which waits for the :attr:`.buffer` to be written to, and then executes the callback
        function on the value provided by the :attr:`.buffer` accessor.

        If the callback is not a coroutine function, it will be run in the default executor pool (see
        https://docs.python.org/3/library/asyncio-eventloop.html#executing-code-in-thread-or-process-pools)

        This method returns a unique identifier (supplied by the :attr:`.token_factory`) which can be used to deregister
        the callback at a later time. Registering a callback multiple times will make the
        callback run multiple times for each published value. One needs to keep the token around to later deregister
        (as you don't get the same token again by calling :meth:`register_callback` with the same callback).

        See Also:

            :attr:`task_manager` -- how to change who is responsible for executing callback tasks

        Args:
            callback: some callable which acts on a TopicDataRecord

        Returns:
            a unique token (supplied by the :attr:`token_factory`) to deregister the callback later
        """
        token = self.token_factory()

        self.callback_tasks[token] = self.task_nursery.create_task(
            TopicCoroutine(buffer=self, callback=callback)
        )

        # callbacks might not be hashable, so we use their id to attach the information
        existing = self._callback_tokens.setdefault(id(callback), ())
        self._callback_tokens[id(callback)] = existing + (token,)
        weakref.finalize(callback, lambda key: self._callback_tokens.pop(key, None), id(callback))

        return token

    async def unregister_callback(self, token: T_Token, timeout=None) -> bool:
        """
        Unregister a callback with the unique token from the registration.
        Also cancel the callback task.

        Args:
            timeout: don't wait indefinitely for task to cancel
            token: from :meth:`.register_callback`

        Returns:
            True if operation succeeded, False if no callback for the token is registered.

        Raises:
            TimeoutError: If ``timeout`` is set and the task does not complete (cancel) in time --
                default behaviour: block until task is canceled successfully
        """
        task = self.callback_tasks.pop(token, None)
        if task is None:
            warnings.warn(f"No callback for {token} found in {self}")
            return False

        stop = asyncio.create_task(self.task_nursery.stop_task(task))
        _, pending = await asyncio.wait((stop,), timeout=timeout)
        return not pending


class BasicTopic(Topic[ubii.proto.TopicDataRecord, int]):
    """
    A :class:`Topic` with a :attr:`~Topic.token_factory` simply producing ascending integers and
    a simple :attr:`.buffer` for :class:`~ubii.proto.TopicDataRecord` messages
    """

    class integer_token_factory:
        """
        creates increasing integers ::

            >>> factory = integer_token_factory()
            >>> another_factory = integer_token_factory()
            >>> factory()
            0
            >>> another_factory()
            1
            >>> factory()
            2
            ...

        """
        __last_token__ = -1

        def __call__(self):
            self.__last_token__ += 1
            return self.__last_token__

    def __init__(self, pattern, **kwargs) -> None:
        super().__init__(pattern, token_factory=self.integer_token_factory(), **kwargs)
        self._buffer_value: ubii.proto.TopicDataRecord | None = None

    def _set_buffer(self, value):
        self._buffer_value = value

    def _get_buffer(self):
        return self._buffer_value

    @util.cached_property
    def buffer(self) -> util.accessor[ubii.proto.TopicDataRecord]:
        """
        You can use the :meth:`~codestare.async_utils.descriptor.accessor.get` and
        :meth:`~codestare.async_utils.descriptor.accessor.set` methods to synchronize access to the internal
        :class:`ubii.proto.TopicDataRecord` value.

        See Also:
           :class:`~codestare.async_utils.descriptor.accessor` -- details about access to shared resource

        """
        return util.accessor(funcs=(self._get_buffer, self._set_buffer), name='buffer')


Topic_co = typing.TypeVar('Topic_co', bound=Topic, covariant=True)


class TopicStore(util.MatchMappingMixin, util.DefaultHookMap[str, Topic_co], typing.Generic[Topic_co]):
    """
    A TopicStore acts like a :class:`~collections.defaultdict` mapping of :math:`pattern \\rightarrow Topic`,
    but allows for complex matching of keys using :meth:`.match_pattern` and :meth:`.match_name`

    Example:

        >>> from ubii.framework.topics import TopicStore
        >>> class Topic:
        ...     def __init__(self, pattern):
        ...             self.pattern = pattern
        ...
        >>> store = TopicStore(default_factory=Topic)
        >>> store.default_factory('topic/glob/pattern_one')
        >>> store.default_factory('topic/glob/pattern_two')
        >>> [topic.pattern for topic in store.match_pattern('topic*')]
        ['topic/glob/pattern_one', 'topic/glob/pattern_two'])
    """

    class on_create_register_callback:
        """
        This decorator can simply be applied once, if you later want to add additional callbacks
        just add them to the :attr:`.callbacks` of the registered :class:`on_create_register_callback`
        """

        def __init__(self, *callbacks):
            self.callbacks = callbacks

        def __call__(self, on_create: util.hook):
            @functools.wraps(on_create)
            def decorated(store: TopicStore, key: str):
                on_create(store, key)
                for cb in self.callbacks:
                    store[key].register_callback(cb)

            return decorated


class StreamSplitRoutine(util.CoroutineWrapper[typing.Any, typing.Any, None]):
    """
    A StreamSplitRoutine splits :class:`TopicDataRecords <ubii.proto.TopicDataRecord>`
    from a :class:`~ubii.proto.TopicData` stream to the buffers of topics from a :class:`TopicStore` container

    Warning:

        This Coroutine only works with Topics with :class:`~ubii.proto.TopicDataRecord` buffers.
        For topics with different buffer type, use an adjusted :class:`StreamSplitRoutine`.
    """

    def __init__(self, *,
                 stream: typing.AsyncIterator[ubii.proto.TopicData],
                 container: TopicStore[Topic[ubii.proto.TopicDataRecord, typing.Any]],
                 logger: logging.Logger | None = None):
        """
        Set the static attributes for the coroutine, the :class:`~ubii.proto.TopicData` source and
        :class:`~ubii.proto.TopicDataRecord` container

        Args:
            stream: source for data
            container: sink for data
            logger: optional logger for processed data
        """

        self._stream: typing.AsyncIterator[ubii.proto.TopicData] = stream
        self._container = container
        self._logger = logger or logging.getLogger(__name__)
        super().__init__(coroutine=self.split_to_topics())

    async def make_record(self) -> typing.AsyncIterator[ubii.proto.TopicDataRecord]:
        """
        Creates iterator for :class:`~ubii.proto.TopicDataRecord` messages from the
        :class:`~ubii.proto.TopicData` iterator passed as ``stream`` during initialization,
        by unpacking possible :attr:`~ubii.proto.TopicData.topic_data_record_list`
        fields

        Returns:
            Iterator over individual records
        """
        async for data in self._stream:
            if data.topic_data_record_list:
                for record in data.topic_data_record_list.elements:
                    yield record
            elif data.topic_data_record:
                yield data.topic_data_record
            else:
                raise data.error

    async def split_to_topics(self) -> None:
        """
        Uses :meth:`.make_record` to create a stream of records, then iterates over them and

            *   creates the topic inside the :class:`TopicStore` passed as ``container`` during initialization
                if no topics matching the glob pattern of received
                :class:`TopicDataRecords <ubii.proto.TopicDataRecord>` are found
            *   sets the topic's :attr:`~Topic.buffer` to the record
        """
        async for record in self.make_record():
            topics = self._container.match_name(record.topic)
            self._logger.log(level=5, msg=f"Record Topic: {record.topic} -> matching: {','.join(map(str, topics))}")
            if not topics:
                topics = (self._container[record.topic],)
                log.warning(f"No topics found for record with topic {record.topic}")

            await asyncio.gather(*[topic.buffer.set(record) for topic in topics or ()])


class OnSubscribeCallback(Protocol):
    async def __call__(self, client_id: str, topic: Topic, as_regex: bool = ..., unsubscribe: bool = ...):
        """
        Callbacks for the :class:OnSubscribersChange` objects need to have this signature

        Args:
            client_id: id of client node
            topic: topic that was (un)subscribed
            as_regex: patterns need to be handled differently from simple strings
            unsubscribe: is this a subscription or a subscription cancellation?
        """


class OnSubscribersChange:
    """
    This callable is a proxy for its :attr:`.callback` (see :class:`OnSubscribeCallback`).

    Attributes:
        as_regex (bool): static argument for eventual call to :attr:`.callback`
        client_id (str): static argument for eventual call to :attr:`.callback`
        event (asyncio.Event): application code can wait for this event -- is set whenever this is called
        callback (OnSubscribeCallback): actual callback
    """

    def __init__(self,
                 client_id: str,
                 as_regex: bool,
                 callback: OnSubscribeCallback):
        self.as_regex: bool = as_regex
        self.client_id: str = client_id
        self.event: asyncio.Event = asyncio.Event()
        self.callback: OnSubscribeCallback = callback

    def __call__(self, topic: Topic, change: typing.Tuple[int, int]) -> asyncio.Task | None:
        """
        Called when subscriber counts change in a topic.

        Executes the :attr:`.callback` when the subscriber count increases to 1 or decreases to 0 -- with args
            *   client_id = :attr:`.client_id`
            *   topic = reference to self
            *   as_regex = :attr:`.as_regex`
            *   unsubscribe = True if subscriber count decreased to 0, False if it was increased to 1

        Args:
            topic: typically the topic which owns this callable
            change: old and new subscriber count

        """
        old, new = change
        task = None

        if new == 0 and old > 0:
            task = topic.task_nursery.create_task(
                self.callback(self.client_id, topic, as_regex=self.as_regex, unsubscribe=True)
            )

        if new == 1 and old < 1:
            task = topic.task_nursery.create_task(
                self.callback(self.client_id, topic, as_regex=self.as_regex, unsubscribe=False)
            )

        if task:
            task.add_done_callback(lambda _: self.event.set())
        else:
            self.event.set()

        return task


class MetaMuxRecord(ubii.proto.TopicDataRecord, metaclass=ubii.proto.ProtoMeta):
    """
    TopicDataRecords need to get additional meta information when they are produced by a Muxer.
    This meta information is not part of the proto schema. This mechanism is similar to the JS way
    muxers ands demuxers are implemented, to keep the processing code comparable.
    """

    __record_meta__: typing.MutableMapping[int, typing.MutableMapping] = {}

    def __init__(self, mapping: typing.Dict | proto.Message = None, *, ignore_unknown_fields: bool = False, **kwargs):
        """

        Args:
            mapping: A dictionary or message to be
                used to determine the values for this message. If it is a dictionary, keys that are not part
                of the proto schema will be added to the metadata instead
            ignore_unknown_fields: If True, do not raise errors for
                unknown fields. Only applied if `mapping` is a mapping type or there
                are keyword parameters.
            **kwargs: Keys and values corresponding to the fields of the message.
                Any keyword arguments that are passed on initialisation, that are not part of
                the :class:`~ubii.proto.TopicDataRecord` schema, will be simply inserted into the metadata mapping.
        """
        field_names = list(ubii.proto.TopicDataRecord.pb().DESCRIPTOR.fields_by_name)

        # let's help super().__init__  a bit by handling unusual behaviours
        if isinstance(mapping, ubii.proto.TopicDataRecord):
            mapping = ubii.proto.TopicDataRecord.pb(mapping)
        elif isinstance(mapping, typing.Mapping):
            kwargs.update(**{k: v for k, v in mapping.items() if k not in field_names})
            mapping = {k: v for k, v in mapping.items() if k in field_names}

        metadata = {k: v for k, v in kwargs.items() if k not in field_names}
        kwargs = {k: v for k, v in kwargs.items() if k in field_names}

        super().__init__(mapping=mapping, ignore_unknown_fields=ignore_unknown_fields, **kwargs)
        self.metadata(**metadata)

    def metadata(self: MetaMuxRecord, **kwargs) -> typing.Optional[typing.MutableMapping]:
        """
        :class:`~ubii.proto.TopicDataRecord` overwrites attribute access, so we can't write any
        attributes that are not part of the protobuf specification.

        This method allows saving metadata for a record in a specified way, without altering
        the proto schema.

        Args:
            **kwargs: If you specify keyword arguments, they will be written to the metadata mapping

        Returns:
            if metadata has been associated with the record, returns the metadata mapping, otherwise returns None
        """
        if kwargs:
            if id(self) not in self.__record_meta__:
                # add finalizer to remove object id when record is garbage collected
                weakref.finalize(self, lambda key: self.__record_meta__.pop(key, None), id(self))
            mapping = self.__record_meta__.setdefault(id(self), {})
            mapping.update(kwargs)

        return self.__record_meta__.get(id(self), None)

    @property
    def type(self) -> str | None:
        """
        Returns attribute name of the attribute that is set in the `type` oneof group for convenience
        """
        return type(self).pb(self).WhichOneof('type')


class TopicMuxer(ubii.proto.TopicMux, metaclass=util.ProtoRegistry):
    """
    A :class:`TopicMuxer` defines a special :attr:`.records` attribute, to handle topic data records. The muxer
    will identify metadata for the records, and associate it with them.
    """
    __unique_key_attr__ = 'id'

    def __init__(self, mapping=None, **kwargs):
        if isinstance(mapping, ubii.proto.TopicMux):
            mapping = ubii.proto.TopicMux.pb(mapping)

        super().__init__(mapping=mapping, **kwargs)

        if not self.id:
            raise ValueError(f"Can't create {type(self)} object without valid 'id'")

        self._records: typing.MutableMapping[typing.Any, TopicMuxer.MuxedRecord] = {}

    def _compute_identity(self, record):
        identity_matches = re.findall(self.identity_match_pattern, record.topic)
        if len(identity_matches) > 0:
            if len(identity_matches) != 1:
                warnings.warn(f"{len(identity_matches)} identity matches found for topic {record.topic}, using first."
                              f" Consider refining the 'identity_match_pattern' regular expression "
                              f"{self.identity_match_pattern!r}")

            return identity_matches[0]

    def _get_records(self) -> 'typing.List[MetaMuxRecord]':
        return list(self._records.values())

    def _identify_records(self, records: typing.List[ubii.proto.TopicDataRecord]):
        for record in records:
            assert isinstance(record, ubii.proto.TopicDataRecord)
            assert 'topic' in record, "Invalid record, no topic set!"

            metarecord = MetaMuxRecord(mapping=record, identity=self._compute_identity(record))
            if metarecord.type and self.data_type != metarecord.type:
                continue

            # we are not updating any records client code might have saved. If you want to get up to date records,
            # use the records condition property
            self._records[record.topic] = metarecord

    records = util.condition_property(fget=_get_records, fset=_identify_records)
    """
    Use this :class:`util.condition_property` to read and write records handled by the
    muxer (which will convert them to :class:`MetaMuxRecord` objects
    that carry additional meta information)

    Example:
        *The example uses an async interpreter* ::
        
            >>> import asyncio
            >>> from ubii.framework.topics import TopicMuxer
            >>> import ubii.proto
            >>> muxer = TopicMuxer(
            ...     id="fake-id",
            ...     data_type='int32',
            ...     topic_selector='/topic/*',
            ...     identity_match_pattern='(?:/topic/([0-9a-z-]+))'
            ... )
            >>> records = [ubii.proto.TopicDataRecord(topic=f"/topic/{num}", int32=num) for num in range(5)]
            >>> records
            [topic: "/topic/0"
            int32: 0
            , topic: "/topic/1"
            int32: 1
            , topic: "/topic/2"
            int32: 2
            , topic: "/topic/3"
            int32: 3
            , topic: "/topic/4"
            int32: 4
            ]
            >>> await muxer.records.set(records)
            >>> muxed = await muxer.records.get(predicate=lambda _ : True)
            >>> muxed[0].metadata()
            {'identity': '0'}
            
    """


class TopicDemuxer(ubii.proto.TopicDemux, metaclass=util.ProtoRegistry):
    """
    A Demuxer converts :class:`MetaMuxRecord` objects to regular records, by setting their attributes
    (currently only the topic) according to their associated metadata and it's own attributes.
    """
    DEFAULT_REGEX_OUTPUT_PARAM = r'{{#([0-9]+)}}'
    __unique_key_attr__ = 'id'

    def __init__(self, mapping=None, param_regex=DEFAULT_REGEX_OUTPUT_PARAM, **kwargs):
        if isinstance(mapping, ubii.proto.TopicDemux):
            mapping = ubii.proto.TopicDemux.pb(mapping)

        super().__init__(mapping=mapping, **kwargs)

        if not self.id:
            raise ValueError(f"Can't create {type(self)} object without valid 'id'")

        self._param_regex = re.compile(param_regex)

    @staticmethod
    def _replace_with_param(params: typing.Tuple | None, match: re.Match):
        # fallback = whole match, no replacements

        fallback = match.group()
        if not len(match.groups()) == 1:
            warnings.warn(f"{len(match.groups())} groups in match, 1 is required")
            return fallback

        try:
            num_param = int(match.group(1))
        except TypeError:
            warnings.warn(f"Can't convert first match group {match.group(1)} to integer")
            # fallback, return whole match, don't replace
            return fallback

        if params is None:
            warnings.warn(f"No parameters specified")
            return fallback

        try:
            param = params[num_param]
        except IndexError:
            warnings.warn(f"Param {num_param} from parameters {params} can't be accessed")
            return fallback

        return param

    def _convert(self, meta_record: MetaMuxRecord | typing.Dict) -> ubii.proto.TopicDataRecord | None:
        if not isinstance(meta_record, MetaMuxRecord):
            meta_record = MetaMuxRecord(mapping=meta_record)

        if meta_record.type and self.data_type != meta_record.type:
            return

        topic = self._param_regex.sub(
            functools.partial(self._replace_with_param, meta_record.metadata().get('output_params')),
            self.output_topic_format
        )

        meta_record.topic = topic or meta_record.topic
        return meta_record

    def convert_record_objects(self, records: typing.List) -> ubii.proto.TopicDataRecordList:
        """
        The Demuxer converts the records to actual TopicDataRecords with the right topic

        Args:
            records: records with metadata, likely created as outputs of a processing module

        Returns:
            elements contain all records that where successfully converted

        Example:
            *The example uses an async interpreter* ::

                >>> import asyncio
                >>> from ubii.framework.topics import TopicDemuxer
                >>> import ubii.proto
                >>> records = [{'int32': num, 'output_params': (str(num),)} for num in range(5)]
                >>> demuxer = TopicDemuxer(
                ...     id='fake-id',
                ...     data_type='int32',
                ...     output_topic_format='/topic/{{#0}}'
                ... )
                >>> demuxer.convert_record_objects(records)
                elements {
                  topic: "/topic/0"
                  int32: 0
                }
                elements {
                  topic: "/topic/1"
                  int32: 1
                }
                elements {
                  topic: "/topic/2"
                  int32: 2
                }
                elements {
                  topic: "/topic/3"
                  int32: 3
                }
                elements {
                  topic: "/topic/4"
                  int32: 4
                }
        """

        return ubii.proto.TopicDataRecordList(elements=[r for r in map(self._convert, records) if r is not None])



import argparse
import ast
import asyncio
import functools
import logging
import typing
import warnings
from pathlib import Path

from ubii.framework.client import UbiiClient

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata

log = logging.getLogger(__name__)

SETUPTOOLS_PM_ENTRYPOINT_KEY = 'ubii.processing_modules'
"""
Processing modules need to register their entry points with this key
"""


def import_name(name: str):
    """
    Utility method, imports the module where the name is located

    Example::

        client_type = import_name('ubii.framework.client.UbiiClient')

    Args:
        name: full name, e.g. 'the.module.attribute'

    Returns:
        the attribute with specified name, from the imported module
    """
    module, attribute = name.rsplit('.', maxsplit=1)
    assert module
    assert attribute
    from importlib import import_module
    module = import_module(module)
    return getattr(module, attribute)


def log_to_folder(log_config, folder='logs/'):
    """
    Changes the log config so that all logs are written to the specified folder
    Args:
        folder: path to folder that should be prefixed
        log_config: configuration dictionary in :func:`logging.config.dictConfig` format

    Returns:
        changed dictionary
    """
    for name, config in log_config['handlers'].items():
        log_file = config.get('filename', None)
        if log_file is None:
            continue

        if not log_file.startswith(folder):
            log_file = f"{folder}{log_file}"

        log_file = Path(log_file)
        config['filename'] = str(log_file)
        log_file.parent.mkdir(parents=True, exist_ok=True)

    return log_config


def load_pm_entry_points() -> typing.Dict[str, typing.Any]:
    """
    Loads setuptools entrypoints for key :attr:`SETUPTOOLS_PM_ENTRYPOINT_KEY`

    Returns:
        list of :class:`~ubii.framework.processing.ProcessingRoutine` types
    """
    with warnings.catch_warnings():
        # this deprecation is discussed a lot
        warnings.simplefilter("ignore")
        entries = [entry for entry in metadata.entry_points().get(SETUPTOOLS_PM_ENTRYPOINT_KEY, ())]
        return {entry.name: entry.load() for entry in entries}


def parse_args():
    """
    Enhanced argument parsing which allows to also load installed processing
    modules and pass arguments to them when they are loaded by the client
    """
    import argparse
    import re
    from ubii.framework.logging import parse_args

    def pm_type(arg_value, pat=r'(?:(.*):)?(.*)'):
        matched = re.match(pat, arg_value)
        if not matched:
            raise argparse.ArgumentTypeError(f"{arg_value} has wrong format, allowed: {pat}")

        return matched.groups()

    kv = r'(?:.*)=(?:.*)'  # a key value pair regex, only capturing the pair

    def mod_arg_type(arg_value, pat=r'(?:(.*):)?({kv}(?:,{kv})*)'.format(kv=kv)):
        matched = re.match(pat, arg_value)
        if not matched:
            raise argparse.ArgumentTypeError(f"{arg_value} has wrong format, allowed: {pat}")

        name, args = matched.groups()
        argdict = {}
        for a in args.split(','):
            key, value = a.rsplit('=', maxsplit=1)
            argdict[key] = ast.literal_eval(value)

        return name, argdict

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--processing-modules', action='append', default=[],
        type=pm_type,
        help='Import processing modules to load. '
             'You can also use the format {name}:{type} to load them with specific names.'
    )
    parser.add_argument(
        '--no-discover', action='store_true', default=False,
        help="Don't use the automatic processing module discovery mechanism"
    )
    parser.add_argument(
        '--module-args', action='append', default=[],
        type=mod_arg_type,
        help="Format {name}:{key}={value}(,{key}={value})*, e.g. my-module:tags=['custom'],description='Wow!'"
    )

    return parse_args(parser=parser)


def create_pm_factories(args: argparse.Namespace) -> typing.Dict[str, typing.Any]:
    pms = {}
    if not args.no_discover:
        pms.update(**load_pm_entry_points())
    if args.processing_modules:
        for name, value in args.processing_modules:
            pms[name] = import_name(value)

    for name, argdict in args.module_args:
        if name not in pms:
            warnings.warn(f"Arguments for module {name} specified but no module with that name was loaded")
            continue

        # update arguments
        pms[name] = functools.partial(pms[name], **argdict)

    if not pms:
        warnings.warn(f"No processing modules imported")
    else:
        print(f"Loaded processing module factories {', '.join(map(repr, pms.values()))}")

    return pms


def main():
    """
    Entry point for cli script see :ref:`CLI` in the documentation
    """
    from ubii.node import connect_client
    from ubii.framework.client import InitProcessingModules
    from codestare.async_utils.nursery import TaskNursery
    from ubii.framework import logging_setup

    pms = create_pm_factories(parse_args())

    log_config = logging_setup.change(
        config=log_to_folder(logging_setup.effective_config.config)
    )

    async def run():
        with connect_client() as client:
            try:
                client.is_dedicated_processing_node = True
                client[InitProcessingModules] = InitProcessingModules(module_factories=pms)
                assert client.implements(InitProcessingModules)
                await client
                while client.state != client.State.UNAVAILABLE:
                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                pass

        await asyncio.sleep(1)
        loop.stop()

    loop = asyncio.get_event_loop_policy().get_event_loop()
    nursery = TaskNursery(name="__main__", loop=loop)

    with log_config:
        nursery.create_task(run())
        loop.run_forever()

    assert not loop.is_running()
    loop.close()


def info_log_client():
    """
    Example for tutorial to create a simple client that prints messages received in the info topics,
    and continuously publishes a value to a custom info topic. See :ref:`Client Example`
    """
    import asyncio
    import argparse
    from ubii.framework.logging import parse_args
    parser = argparse.ArgumentParser()
    parser.add_argument('--url', type=str, action='store', default=None, help='URL of master node service endpoint')
    args = parse_args(parser=parser)

    from ubii.node import connect_client, Subscriptions, Publish, LegacyProtocol

    loop = asyncio.get_event_loop_policy().get_event_loop()

    async def run():
        client: UbiiClient[LegacyProtocol]
        async with connect_client(url=args.url) as client:
            # we don't hard code the topic, we use the DEFAULT TOPIC from the master node
            constants = client.protocol.context.constants
            assert constants
            info_topic_pattern = constants.DEFAULT_TOPICS.INFO_TOPICS.REGEX_ALL_INFOS
            info, = await client[Subscriptions].subscribe_regex(info_topic_pattern)

            print(f"Subscribed to topic {info_topic_pattern!r}")
            info.register_callback(print)

            value = None

            while client.state != client.State.UNAVAILABLE:
                value = 'foo' if value == 'bar' else 'bar'
                await asyncio.sleep(1)
                await client[Publish].publish({'topic': '/info/custom_topic', 'string': value})

        loop.stop()

    from codestare.async_utils.nursery import TaskNursery
    nursery = TaskNursery(name="__main__", loop=loop)
    nursery.create_task(run())
    loop.run_forever()

    assert not loop.is_running()
    loop.close()


if __name__ == '__main__':
    main()



import asyncio

import pytest

from ubii.framework.client import Subscriptions, Publish

pytestmark = pytest.mark.asyncio


class TestSubscriptions:

    @pytest.fixture(scope='class', autouse=True)
    async def startup(self, client):
        await client
        assert client.implements(Subscriptions, Publish)

    @pytest.mark.parametrize('test_params', ({'a': 1, 'b': 2, 'c': 3},))
    async def test_basic_subscriptions(self, client, test_params):
        topics = await client[Subscriptions].subscribe_topic(*(f"{client.id}/{name}" for name in test_params))

        for topic in topics:
            _, suffix = topic.pattern.rsplit('/', maxsplit=1)
            await client[Publish].publish({'topic': topic.pattern, 'int32': test_params[suffix]})

        received = await asyncio.gather(*[top.buffer.get(predicate=lambda v: v is not None) for top in topics])
        for received, sent in zip(received, test_params.values()):
            assert received.int32 == sent

    @pytest.mark.parametrize('pattern, value', [('foo*', 1), ])
    async def test_regex_subscriptions(self, client, pattern, value):
        topic, = await client[Subscriptions].subscribe_regex(f"{client.id}/{pattern}")

        await client[Publish].publish({'topic': topic.pattern[:-1] + 'bar', 'int32': value})
        received = await topic.buffer.get(predicate=lambda v: v is not None)

        assert received.int32 == value
